Require Import Relations.
Require Import Equality.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import vgtac.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import Events.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import SemanticsLTS.
Require Import SemanticsC.
Require Import SemanticsCLTS.
Require Import MemoryRelationC.
Require Import Behaviors.
Require Import Simulation.
Require Import SimulationLocalC.

Import SyntaxNotations.

Section Adequacy_General.
  Variable lts_src: @LTS ltst_lang.
  Variable lts_tgt : @LTS ltstC_lang.
  
  Hypothesis lts_src_property : lts_propertyT lts_src.
  Hypothesis lts_tgt_property : ltsC_propertyT lts_tgt.

(*  Hypothesis error_sim : forall scfg_src scfg_tgt stack_src stack_tgt mrel_init mrel_entry mrel i cfg_src cfg_tgt
    (Hsim_functions: sim_local_functions lts_src lts_tgt scfg_src scfg_tgt mrel_init)
    (Hle : MemRel.le mrel_init mrel)
    (Hsim : sim_local lts_src lts_tgt scfg_src scfg_tgt stack_src stack_tgt mrel_entry mrel i cfg_src cfg_tgt)
    (Hnormal_src : get_status lts_src cfg_src = Status.normal)
    (Hnormal_tgt : get_status lts_tgt cfg_tgt = Status.normal)
    cfg'_tgt (Hstep_tgt : taustep lts_tgt scfg_tgt cfg_tgt cfg'_tgt)
    (Herror_tgt : get_status lts_tgt cfg'_tgt = Status.error),
    exists cfg'_src, rtc (taustep lts_src scfg_src) cfg_src cfg'_src /\ get_status lts_src cfg'_src = Status.error.*)

  Inductive sim_local_seq_ scfg_src scfg_tgt (mrel: MemRelC.t) (i: WF.t) : option Semantics.State -> option State -> Prop :=
  | sim_global_base: forall st_src st_tgt
      (Hinstrs_src: st_src.(Semantics.context).(Semantics.instrs) = nil)
      (Hstack_src: st_src.(Semantics.stack) = nil)
      (Hinstrs_tgt: st_tgt.(context).(instrs) = nil)
      (Hstack_tgt: st_tgt.(stack) = nil),
    sim_local_seq_ scfg_src scfg_tgt mrel i (Some st_src) (Some st_tgt)
  | sim_global_step: forall context'_src context'_tgt stack'_src stack'_tgt mrel' cfg_src cfg_tgt
      (SIM_local: sim_local lts_src lts_tgt scfg_src scfg_tgt 
                    (context'_src::stack'_src) (context'_tgt::stack'_tgt) mrel' mrel i cfg_src cfg_tgt)
      (SIM_global: forall mrel'_future st'_src st'_tgt 
           (Hnormal_src: get_status lts_src (Some st'_src) <> Status.error)
           (Hnormal_tgt: get_status lts_tgt (Some st'_tgt) <> Status.error)
           (Hmrel'_future_le: MemRelC.le_public mrel' mrel'_future)
           (Hst': mrel'_future st'_src.(Semantics.mem) st'_tgt.(mem))
           (Hinstrs'_src: st'_src.(Semantics.context) = context'_src)
           (Hstack'_src: st'_src.(Semantics.stack) = stack'_src)
           (Hinstrs'_tgt: st'_tgt.(context) = context'_tgt)
           (Hstack'_tgt: st'_tgt.(stack) = stack'_tgt),
         exists i',
         sim_local_seq_ scfg_src scfg_tgt mrel'_future i' (Some st'_src) (Some st'_tgt)),
    sim_local_seq_ scfg_src scfg_tgt mrel i cfg_src cfg_tgt
  .

  Definition sim_local_seq scfg_src scfg_tgt mrel_init i st_src st_tgt := 
    exists mrel, 
      MemRelC.le mrel_init mrel /\
      sim_local_seq_ scfg_src scfg_tgt mrel i st_src st_tgt.

  Lemma rtc_cases: forall {A} (P : Prop) R (x z : A), rtc R x z -> (forall y, R x y -> rtc R y z -> P) ->
    (x = z -> P) -> P.
  Proof.
    intros ? ? ? ? ? H; induction H; eauto; intros.
    apply IHclos_refl_trans1; intros.
    - apply (H1 y0); auto.
      eapply rt_trans; eauto.
    - subst; apply IHclos_refl_trans2; eauto.
  Qed.

  Lemma sim_local_seq_is_simulation: 
    forall scfg_src scfg_tgt mrel_init
      (Hsim_functions: sim_local_functions lts_src lts_tgt scfg_src scfg_tgt mrel_init),
    sim_local_seq scfg_src scfg_tgt mrel_init <3= sim _ _ lts_src lts_tgt scfg_src scfg_tgt.
  Proof.
    intros ? ? ? ?; pcofix CIH.
    intros; pfold.
    inversion PR.
    destruct H as [Hmrel_init Hsim_local]; inversion Hsim_local; simtac.
    - eapply sim_term; eauto.
      + destruct lts_src_property.
        eapply lts_term; eauto.
        by unfold Semantics.is_terminated; rewrite Hinstrs_src, Hstack_src.
      + destruct lts_tgt_property.
        eapply lts_term; eauto.
        by unfold is_terminated; rewrite Hinstrs_tgt, Hstack_tgt.
    - pinversion SIM_local; simtac.
      + eapply sim_error; eauto.
      + apply sim_step; auto.
        intros; eexists; split; eauto.
        destruct lts_tgt_property.
        exploit lts_return; eauto; simtac.
        { rewrite Hinstrs_tgt; auto. }
        { instantiate (5 := Some _); unfold semC_instr; rewrite Hinstrs_tgt; rewrite Hstack_tgt; simtac. }
        intro X; destruct X as [tgt_step tgt_status].
        exploit (lts_return_deterministic scfg_tgt (Some st_tgt)).
        { simpl; rewrite Hinstrs_tgt; auto. }
        { eauto. }
        { apply Htgt. }
        intros; simtac.
        destruct lts_src_property.
        exploit (lts_return0 scfg_src (Some st2_src)); eauto; simtac.
          by rewrite Hinstrs_src; auto.
          instantiate (3 := Some _); unfold sem_instr; rewrite Hinstrs_src, Hstack_src; done.
        i; des.
        exploit SIM_global; eauto; intro; simtac.
        exists x4; eexists; split.
        apply indexed_step_step; apply lts_return0; simtac.
        rewrite Hinstrs_src; auto.
        instantiate (1 := Some _); unfold sem_instr; rewrite Hinstrs_src; rewrite Hstack_src; simtac.
        right; eapply CIH; eauto.
        unfold sim_local_seq; exists mrel2; split; eauto.
        etransitivity; eauto.
      + destruct ((Semantics.fenv scfg_src) ! calleefid) eqn: fid_src.
        destruct (Semantics.bind_params (FDecl.params t) argvals_src st2_src.(Semantics.mem) Env.empty) eqn: params_src.
        exploit Hsim_functions; eauto; intro X.
        destruct X as [t_tgt [fid_tgt [Hparams_tgt Hsim_func]]]; vgtac.des.
        inversion lts_tgt_property.
        destruct (bind_params (FDecl.params t_tgt) argvals_tgt st_tgt.(mem) EnvC.empty) eqn: params_tgt.
        exploit (lts_call scfg_tgt); eauto; simtac.
        { rewrite Hinstrs_tgt; auto. }
        { instantiate (3 := Some _); unfold semC_instr; rewrite Hinstrs_tgt; rewrite fid_tgt; simtac.
          eapply bindPP_right; eauto.
          rewrite params_tgt; simtac.
        }
        intro X; destruct X as [tgt_step tgt_status].
        inversion lts_src_property.
        exploit (lts_call0 scfg_src); eauto; simtac.
        { rewrite Hinstrs_src; auto. }
        { instantiate (3 := Some _); unfold sem_instr; rewrite Hinstrs_src; rewrite fid_src; simtac.
          eapply bindPP_right; eauto.
          rewrite params_src; simtac.
        }
        intro X; destruct X as [src_step src_status].
        destruct src_status; simtac.
        destruct tgt_status; simtac.
        apply sim_step; auto.
        intros; eexists; split; eauto.
        exploit (lts_call_deterministic scfg_tgt (Some st_tgt)).
        { simpl; rewrite Hinstrs_tgt; auto. }
        { eauto. }
        { apply tgt_step. }
        intros; simtac.
        exploit Hsim_func; try (etransitivity; eauto); eauto; simtac.
        intro X; destruct X as [i Hsim]; des.
        exists i; eexists; split.
        apply indexed_step_step; apply src_step.
        right; eapply CIH; eauto.
        unfold sim_local_seq.
        exists mrel2; split.
        etransitivity; eauto.
        eapply sim_global_step; eauto; simtac; intros.
        destruct (get_status lts_src (Some st'_src)) eqn: status'_src; simtac.
        { rewrite <- lts_term0 in status'_src; unfold Semantics.is_terminated in status'_src; simtac.
          rewrite Hstack'_src in H2; destruct (Semantics.stack st2_src); simtac. }
        destruct (get_status lts_tgt (Some st'_tgt)) eqn: status'_tgt; simtac.
        { rewrite <- lts_term in status'_tgt; unfold is_terminated in status'_tgt; simtac.
          rewrite Hstack'_tgt in H2; destruct (stack st_tgt); simtac. }
        exploit Hreturn; eauto; intro; simtac.
        exists x2; eapply sim_global_step; eauto.
        { exploit StateRelationC.mrel_bind_params_types; eauto.
          apply StateRelationC.mrel_Env_empty.
          intro; simtac.
          rewrite H in params_tgt; simtac.
        }
        { eapply sim_error.
          eapply rt_trans.
          eauto.
          apply rt_step; inversion lts_src_property.
          apply lts_call; eauto; simtac.
          rewrite Hinstrs_src; auto.
          instantiate (1 := None); unfold sem_instr; rewrite Hinstrs_src; rewrite fid_src; simtac.
          split; eauto.
          eapply bindPP_right; eauto.
          rewrite params_src; simtac.
          apply lts_src_property.
        }
        { eapply sim_error.
          eapply rt_trans.
          eauto.
          apply rt_step; inversion lts_src_property.
          apply lts_call; eauto; simtac.
          rewrite Hinstrs_src; auto.
          instantiate (1 := None); unfold sem_instr; rewrite Hinstrs_src; rewrite fid_src; simtac.
          apply lts_src_property.
        }
      + apply sim_step; auto.
        intros.
        exploit Hstep; eauto; intro; simtac.
        eexists; split; eauto.
        eexists; eexists; split; eauto.
        right; eapply CIH.
        eauto.
        unfold sim_local_seq; exists x6; split.
        etransitivity; eauto.
        eapply sim_global_step; eauto.
  Qed.

  Lemma sim_prog_adequacy:
    forall prog_src prog_tgt
      (Hsim_local_prog: sim_prog lts_src lts_tgt prog_src prog_tgt),
    behave_prog lts_tgt prog_tgt <1= behave_prog lts_src prog_src.
  Proof.
    destruct lts_src_property, lts_tgt_property.
    i; rdes PR.
    specialize (Hsim_local_prog _ _ PR); des.
    r; esplits; eauto.
    eapply lts_load_prog in Hprog_src; des.
    eapply lts_load_prog0 in PR; des; subst.
    destruct (classic 
      (exists main, 
          <<Hmain: scfg_src.(Semantics.fenv) ! (%f "main") = Some main>>
       /\ <<Hmain_params: main.(FDecl.params) = []>>)).
    { des.
      assert (Hsim_main:= Hsim_functions (%f"main") _ Hmain).
      des; subst; dependent destruction Hmrel_init_mem.
      inv Hsim_local_func.
      destruct (func_tgt.(FDecl.params)) eqn: Htgt_params.
      2: rewrite Hmain_params in *; inv Hparams.
      clear Hparams.

      assert (Hcall_state_src: SemanticsLTS.is_call_state (Some st)).
      { by unfold SemanticsLTS.is_call_state; rewrite Hmain_instrs. }

      assert (Hcall_state_tgt: is_call_state (Some st0)).
      { by unfold is_call_state; rewrite Hmain_instrs0. }

      exploit (lts_call0 scfg); eauto.
      { instantiate (1:= Some _).
        r; r; rewrite Hmain_instrs0, Hfdef_tgt.
        s; econs 2; [done|].
        by rewrite Htgt_params. }
      intros [Hstep_tgt Hstatus_tgt].
      destruct Hstatus_tgt; try done.

      exploit (behave_tau_deterministic lts_tgt); eauto.
      clear PR0; intro Hbehave_tgt.

      exploit (lts_call scfg_src); eauto.
      { instantiate (1:= Some _).
        r; r; rewrite Hmain_instrs, Hmain.
        s; econs 2; [done|].
        by rewrite Hmain_params. }
      intros [Hstep_src Hstatus_src].
      destruct Hstatus_src; try done.

      eapply behave_taus, rt_step; eauto.

      rr in Hsim.
      hexploit Hsim; eauto; try eapply list_forall2_nil; try done.
      { reflexivity. }
      { by rewrite Hmain_params. }
      { by rewrite Htgt_params. }
      i; des.

      eapply sim_adequacy; eauto.
      eapply sim_local_seq_is_simulation; eauto.
      r; esplits; [reflexivity|].
      econs 2; try done; eauto.

      i; exists i.
      econs 1; eauto.
      - by rewrite Hinstrs'_src. 
      - by rewrite Hstack'_src, Hmain_stack.
      - by rewrite Hinstrs'_tgt.
      - by rewrite Hstack'_tgt, Hmain_stack0.
      - eauto.
    }
    {
      pfold; econs. 
      - instantiate (1:= None).
        econs; eapply lts_call; eauto.
        + by unfold SemanticsLTS.is_call_state; rewrite Hmain_instrs.
        + split; eauto.
          r; rewrite Hmain_instrs.
          destruct ((Semantics.fenv scfg_src) ! (%f "main")); s; [|done].
          econs 2; [done|].
          destruct (t.(FDecl.params)) eqn: Def_t; [|done]. 
          by exfalso; eauto.
      - by eauto.
    }
  Qed.

End Adequacy_General.

Theorem adequacyC: 
    forall prog_src prog_tgt
      (Hsim_local_prog: sim_prog lts ltsC prog_src prog_tgt),
    behave_prog ltsC prog_tgt <1= behave_prog lts prog_src.
Proof.
  eauto using sim_prog_adequacy, lts_property, ltsC_property.
Qed.
