Require Import Relations.
Require Import Equality.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import vgtac.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import Events.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import Behaviors.
Require Import SimulationLocal.

Import SyntaxNotations.

Definition is_return_state (ost: option State) : bool :=
  match ost with None => false | Some st =>
    match st.(context).(instrs) with
    | nil => true
    | _ => false
    end 
  end.

Definition is_call_state (ost: option State) : bool :=
  match ost with None => false | Some st =>
    match st.(context).(instrs) with
    | (Instr.call _ _) :: _ => true
    | _ => false
    end 
  end.

Inductive lts_propertyT (lts : @LTS ltst_lang) : Prop :=
_lts_propertyT
  (lts_load_prog: 
    forall prg scfg cfg
      (Hload: lts.(Behaviors.load_prog) prg scfg cfg),
    exists st,
    <<Hnormal: lts.(get_status) cfg = Status.normal>> /\
    <<Hcfg: cfg = Some st>> /\
    <<Hmain_instrs: st.(context).(instrs) = [CALL %f"main" WITH []]>> /\
    <<Hmain_stack: st.(stack) = nil>>)
  (lts_error: 
    lts.(get_status) None = Status.error)
  (lts_term: 
    forall st,
    is_terminated st <-> lts.(get_status) (Some st) = Status.term)
  (lts_call: 
    forall scfg cfg evt cfg'
      (Norm: lts.(get_status) cfg = Status.normal)
      (Call: is_call_state cfg)
      (Step: sem_step sem_instr scfg cfg evt cfg'),
    lts.(step) scfg cfg evt cfg' /\ (cfg' = None \/ lts.(get_status) cfg' = Status.normal))
  (lts_call_deterministic: 
    forall scfg cfg
      (Call: is_call_state cfg),
    deterministic lts scfg cfg)
  (lts_return: 
    forall scfg cfg evt cfg'
      (Norm: lts.(get_status) cfg = Status.normal)
      (Call: is_return_state cfg)
      (Step: sem_step sem_instr scfg cfg evt cfg'),
    lts.(step) scfg cfg evt cfg' /\ lts.(get_status) cfg' <> Status.error)
  (lts_return_deterministic: 
    forall scfg cfg
      (Call: is_return_state cfg),
    deterministic lts scfg cfg)
.

Lemma lts_property: lts_propertyT lts.
Proof.
  split; simtac.
  - unfold Semantics.load_prog; intros; simtac.
    inv Hload. clear Hload. destruct Hr2. subst. simpl in *.
    eexists. splits; eauto.
  - intro; destruct (is_terminated st) eqn: term; simtac.
  - intros; split; auto.
    destruct cfg; simtac.
    destruct (instrs (context s)) eqn: Hinstrs; simtac.
    destruct cfg'; [right; simtac | left; auto].
    destruct t; simtac.
    unfold sem_instr in Step; rewrite Hinstrs in Step; simtac.
  - intros; destruct cfg; simtac.
    destruct (instrs (context s)) eqn: Hinstrs; simtac.
    destruct t; simtac.
    unfold deterministic; intros; simtac.
    unfold sem_instr in *; rewrite Hinstrs in STEP1; rewrite Hinstrs in STEP2; simtac.
    destruct cfg1; destruct cfg2; simtac.
    + rewrite Hr0 in Hr1; simtac; rewrite Hcond0 in Hcond1; simtac.
    + inversion H; simtac.
      * rewrite Hr0 in Hr1; simtac.
      * rewrite Hr0 in Hr1; simtac; rewrite Hcond0 in Hr2; simtac.
    + inversion H; simtac.
      * rewrite Hr0 in Hr1; simtac.
      * rewrite Hr0 in Hr1; simtac; rewrite Hcond0 in Hr2; simtac.
  - intros; split; auto.
    destruct cfg; simtac.
    destruct (instrs (context s)) eqn: Hinstrs; simtac.
    unfold sem_instr in Step; rewrite Hinstrs in Step; simtac.
    unfold is_terminated in Norm; rewrite Hinstrs in Norm; simtac.
    destruct (stack s); simtac.
    destruct cfg'; simtac.
    destruct (is_terminated {| context := c; stack := l; mem := mem s |}); simtac.
  - intros; destruct cfg; simtac.
    destruct (instrs (context s)) eqn: Hinstrs; simtac.
    unfold deterministic; intros; simtac.
    unfold sem_instr in *; rewrite Hinstrs in STEP1; rewrite Hinstrs in STEP2; simtac.
    destruct (stack s); simtac; destruct cfg1; simtac; destruct cfg2; simtac.
Qed.
