Require Import Relations.
Require Import Equality.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import vgtac.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import MemoryC.
Require Import Events.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import SemanticsC.
Require Import MemoryRelationC.
Require Import StateRelationC.
Require Import Behaviors.

Import SyntaxNotations.

Inductive lift_mrel (mrel:MemRelC.t): forall (cfg1:option Semantics.State) (cfg2:option State), Prop :=
| lift_mrel_
    st1 st2
    (H: mrel st1.(Semantics.mem) st2.(mem)):
    lift_mrel mrel (Some st1) (Some st2)
.

Section SimulationLocal.
  Variable lts_src: @LTS ltst_lang.
  Variable lts_tgt: @LTS ltstC_lang.

  Section Def.
  Variable (scfg_src:Semantics.StaticState).
  Variable (scfg_tgt:StaticState).

  Section SimLocalDef.
  Variable (stack_entry_src:list Semantics.Context).
  Variable (stack_entry_tgt:list Context).
  Variable (mrel_entry:MemRelC.t).

  Inductive sim_local_
            (sim_local: MemRelC.t -> WF.t -> option Semantics.State -> option State -> Prop)
            (mrel:MemRelC.t) (i:WF.t) (cfg_src:option Semantics.State) (cfg_tgt:option State): Prop :=
  | sim_local_error
      cfg2_src (Hst2_src: rtc (lts_src.(taustep) scfg_src) cfg_src cfg2_src)
      (Herror: lts_src.(get_status) cfg2_src = Status.error)
  | sim_local_return
      st2_src (Hst2_src: rtc (lts_src.(taustep) scfg_src) cfg_src (Some (st2_src)))
      (Hnormal_src: lts_src.(get_status) (Some (st2_src)) = Status.normal)
      (Hnormal_tgt: lts_tgt.(get_status) cfg_tgt = Status.normal)
      st_tgt (Hst_tgt: cfg_tgt = Some st_tgt)
      (mrel2:MemRelC.t) (Hmrel2: mrel2 st2_src.(Semantics.mem) st_tgt.(mem))
      (Hinstrs_src: st2_src.(Semantics.context).(Semantics.instrs) = nil)
      (Hinstrs_tgt: st_tgt.(context).(instrs) = nil)
      (Hstack_src: st2_src.(Semantics.stack) = stack_entry_src)
      (Hstack_tgt: st_tgt.(stack) = stack_entry_tgt)
      (Hmrel2_le: MemRelC.le mrel mrel2)
      (Hmrel2_le_public: MemRelC.le_public mrel_entry mrel2)
  | sim_local_call
      st2_src (Hst2_src: rtc (lts_src.(taustep) scfg_src) cfg_src (Some st2_src))
      (Hnormal_src: lts_src.(get_status) (Some (st2_src)) = Status.normal)
      (Hnormal_tgt: lts_tgt.(get_status) cfg_tgt = Status.normal)
      st_tgt (Hst_tgt: cfg_tgt = Some st_tgt)
      (mrel2:MemRelC.t) (Hmrel2: mrel2 st2_src.(Semantics.mem) st_tgt.(mem))
      (Hmrel2_le: MemRelC.le mrel mrel2)
      calleefid args_src args_tgt istrs_src istrs_tgt
      (Hinstrs_src: st2_src.(Semantics.context).(Semantics.instrs) = (Instr.call calleefid args_src)::istrs_src)
      (Hinstrs_tgt: st_tgt.(context).(instrs) = (Instr.call calleefid args_tgt)::istrs_tgt)
      (Hstack_src: st2_src.(Semantics.stack) <> nil)
      (Hstack_tgt: st_tgt.(stack) <> nil)
      argvals_tgt (Hargs_tgt: semC_bexp_list scfg_tgt.(genv) st_tgt.(context).(lenv) st_tgt.(mem) args_tgt = Some argvals_tgt)
      argvals_src (Hargs_src: sem_bexp_list scfg_src.(Semantics.genv) st2_src.(Semantics.context).(Semantics.lenv) args_src = Some argvals_src)
      (Hargvals: list_forall2
                   (MemEquivC.rel_value mrel2 st2_src.(Semantics.mem) st_tgt.(mem))
                   argvals_src argvals_tgt)
      (Hreturn:
         forall  mrel3 st3_src st3_tgt
           (Hnormal_src: lts_src.(get_status) (Some (st3_src)) = Status.normal)
           (Hnormal_tgt: lts_tgt.(get_status) (Some (st3_tgt)) = Status.normal)
           (Hmrel3_le: MemRelC.le_public mrel2 mrel3)
           (Hst3_mem: mrel3 st3_src.(Semantics.mem) st3_tgt.(mem))
           (Hst3_context_src: st3_src.(Semantics.context) = Semantics.mkContext st2_src.(Semantics.context).(Semantics.lenv) istrs_src)
           (Hst3_stack_src: st3_src.(Semantics.stack) = st2_src.(Semantics.stack))
           (Hst3_context_tgt: st3_tgt.(context) = mkContext st_tgt.(context).(lenv) istrs_tgt)
           (Hst3_stack_tgt: st3_tgt.(stack) = st_tgt.(stack)),
         exists i3,
         sim_local mrel3 i3 (Some st3_src) (Some st3_tgt))
  | sim_local_step
      (Hnormal: lts_tgt.(get_status) cfg_tgt = Status.normal)
      (Hstep:
         forall evt cfg3_tgt (Htgt: lts_tgt.(step) scfg_tgt cfg_tgt evt cfg3_tgt),
           exists cfg2_src,
             rtc (lts_src.(taustep) scfg_src) cfg_src cfg2_src /\
             exists i3 cfg3_src,
               lts_src.(indexed_step) scfg_src (Indexed.mk i cfg2_src) evt (Indexed.mk i3 cfg3_src) /\
               exists (mrel3:MemRelC.t),
                 MemRelC.le mrel mrel3 /\
                 lift_mrel mrel3 cfg3_src cfg3_tgt /\
                 sim_local mrel3 i3 cfg3_src cfg3_tgt)
  .
  Hint Constructors sim_local_.

  Definition sim_local: _ -> _ -> _ -> _ -> Prop :=
    paco4 sim_local_ bot4.

  Lemma sim_local_mon: monotone4 sim_local_.
  Proof.
    repeat intro; destruct IN; eauto.
    - eapply sim_local_call; eauto.
      intros. exploit Hreturn; eauto.
      intros [i3 Hsim]. exists i3. auto.
    - eapply sim_local_step; eauto.
      intros. exploit Hstep; eauto.
      intros [cfg2_src [Hcfg2_src [i3 [cfg3_src [Hcfg3_src [mrel3 [Hmrel3 [Hmrel4 Hsim]]]]]]]].
      eauto 10.
  Qed.
  End SimLocalDef.
  Hint Constructors sim_local_.
  Hint Resolve sim_local_mon: paco.

 Definition sim_local_func_aux
    mrel_init func_src func_tgt: Prop :=
    forall
      mrel_entry
      (Hmrel_init: MemRelC.le mrel_init mrel_entry)
      stack_src stack_tgt mem_src mem_tgt
      (Hmrel_entry: mrel_entry mem_src mem_tgt)
      (Hstack_entry_src: stack_src <> nil)
      (Hstack_entry_tgt: stack_tgt <> nil)
      argvals_src argvals_tgt
      (Hargvals:
         list_forall2
           (MemEquivC.rel_value mrel_entry mem_src mem_tgt)
           argvals_src argvals_tgt)
      lenv_src (Hlenv_src: Some lenv_src = Semantics.bind_params func_src.(FDecl.params) argvals_src mem_src Env.empty)
      lenv_tgt (Hlenv_tgt: Some lenv_tgt = bind_params func_tgt.(FDecl.params) argvals_tgt mem_tgt EnvC.empty)
      cfg_src
      (Hcfg_src: cfg_src = Some (Semantics.mkState (Semantics.mkContext lenv_src (Instr.initialize func_src.(FDecl.locals)::func_src.(FDecl.body))) stack_src mem_src))
      (Hnormal_src: lts_src.(get_status) cfg_src = Status.normal)
      cfg_tgt
      (Hcfg_tgt: cfg_tgt = Some (mkState (mkContext lenv_tgt (Instr.initialize func_tgt.(FDecl.locals)::func_tgt.(FDecl.body))) stack_tgt mem_tgt))
      (Hnormal_tgt: lts_tgt.(get_status) cfg_tgt = Status.normal),
      exists i,
        <<Hsim: sim_local stack_src stack_tgt mrel_entry mrel_entry i cfg_src cfg_tgt>>.

  Inductive sim_local_func mrel_init func_src func_tgt: Prop :=
  | sim_local_func_
      (Hparams: list_forall2 (fun v_src v_tgt => fst v_src = fst v_tgt) func_src.(FDecl.params) func_tgt.(FDecl.params))
      (Hsim: sim_local_func_aux mrel_init func_src func_tgt)
 .

  Definition sim_local_functions mrel_init :=
    forall fid func_src (Hfunc_src: PTree.get fid scfg_src.(Semantics.fenv) = Some func_src),
    exists func_tgt,
    <<Hfdef_tgt: PTree.get fid scfg_tgt.(fenv) = Some func_tgt>> /\
    <<Hsim_local_func: sim_local_func mrel_init func_src func_tgt>>.
  End Def.
  Hint Constructors sim_local_.
  Hint Resolve sim_local_mon: paco.

  Definition sim_prog prog_src prog_tgt : Prop :=
    forall scfg_tgt cfg_init_tgt 
      (Hprog_tgt: lts_tgt.(load_prog) prog_tgt scfg_tgt cfg_init_tgt),
    exists scfg_src cfg_init_src mrel_init,
    <<Hprog_src: lts_src.(load_prog) prog_src scfg_src cfg_init_src>> /\
    <<Hmrel_init_mem: lift_mrel mrel_init cfg_init_src cfg_init_tgt>> /\
    <<Hmrel_init_genv: mrel_genv mrel_init scfg_src.(Semantics.genv) scfg_tgt.(genv)>> /\
    <<Hsim_functions: sim_local_functions scfg_src scfg_tgt mrel_init>>.

End SimulationLocal.
Hint Constructors sim_local_.
Hint Resolve sim_local_mon: paco.

Ltac simtac :=
  repeat
    (simpl in *; unfold Env.get, SomeP, NoneP, singletonP in *; subst; pclearbot;
     (try done);
     (try basictac_aux);
     (try destac_aux);
     (try monadtac_aux);
     (try maptac_aux);
     (try memtac_aux)).

Lemma error_sim : forall scfg_src scfg_tgt stack_src stack_tgt mrel_init mrel_entry mrel i cfg_src cfg_tgt
    (Hsim_functions: sim_local_functions lts ltsC scfg_src scfg_tgt mrel_init)
    (Hle : MemRelC.le mrel_init mrel)
    (Hsim : sim_local lts ltsC scfg_src scfg_tgt stack_src stack_tgt mrel_entry mrel i cfg_src cfg_tgt)
    (Hnormal_src : get_status lts cfg_src = Status.normal)
    (Hnormal_tgt : get_status ltsC cfg_tgt = Status.normal)
    cfg'_tgt (Hstep_tgt : taustep ltsC scfg_tgt cfg_tgt cfg'_tgt)
    (Herror_tgt : get_status ltsC cfg'_tgt = Status.error),
    exists cfg'_src, rtc (taustep lts scfg_src) cfg_src cfg'_src /\ get_status lts cfg'_src = Status.error.
Proof.
  intros; simtac.
  destruct cfg'_tgt; simtac.
  { destruct (is_terminated s); simtac. }
  exists None; destruct cfg_tgt; simtac; destruct cfg_src; split; simtac.
  pinversion Hsim; simtac.
  - destruct cfg2_src; simtac.
    destruct (Semantics.is_terminated s1); simtac.
  - unfold semC_instr in H; rewrite Hinstrs_tgt in H; simtac.
    destruct (stack st_tgt); simtac.
  - unfold semC_instr in H; rewrite Hinstrs_tgt in H; simtac.
    eapply rt_trans; [eauto | apply rt_step]; simtac; split; auto.
    unfold sem_instr; rewrite Hinstrs_src; simtac.
    destruct ((Semantics.fenv scfg_src) ! calleefid) eqn: fun_src; simtac.
    exploit Hsim_functions; eauto.
    intro X; destruct X as [t_tgt [fun_tgt [Hparams Hsim_fun]]]; des.
    rewrite fun_tgt in H; simtac.
    eapply bindPP_right; eauto.
    destruct (Semantics.bind_params (FDecl.params t) argvals_src (Semantics.mem st2_src) Env.empty) eqn: params_src; simtac.
    rewrite Hargs_tgt in H; inversion H; simtac.
    exploit mrel_bind_params_types; eauto.
    apply mrel_Env_empty.
    intro X; destruct X as [env_tgt params_tgt]; rewrite params_tgt in Hr2; simtac.
  - exploit Hstep.
    instantiate (3 := None); simtac.
    intro; simtac.
    inversion H3.
Qed.
