Require Import Relations.
Require Import Axioms.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import Integers.
Require Import Maps.
Require Import WFType.
Require Import Monad.
Require Import Events.
Require Import Behaviors.

Ltac tac :=
  repeat
    (simpl in *; subst; pclearbot;
     (try done);
     (try basictac_aux);
     (try destac_aux);
     (try monadtac_aux);
     (try maptac_aux)).

Section Simulation.
  Variable
    (ltst_src ltst_tgt: LTS_type)
    (lts_src:@LTS ltst_src)
    (lts_tgt:@LTS ltst_tgt)
    (scfg_src:ltst_src.(StaticConfig))
    (scfg_tgt:ltst_tgt.(StaticConfig)).

  Inductive _sim sim (i:WF.t) (cfg_src:ltst_src.(Config)) (cfg_tgt:ltst_tgt.(Config)): Prop :=
  | sim_error
      cfg2_src (Hst2_src: rtc (lts_src.(taustep) scfg_src) cfg_src cfg2_src)
      (Hsrc: lts_src.(get_status) cfg2_src = Status.error)
  | sim_term
      cfg2_src (Hcfg2_src: rtc (lts_src.(taustep) scfg_src) cfg_src cfg2_src)
      (Hsrc: lts_src.(get_status) cfg2_src = Status.term)
      (Htgt: lts_tgt.(get_status) cfg_tgt = Status.term)
  | sim_step
      (Htgt: lts_tgt.(get_status) cfg_tgt = Status.normal)
      (Hstep:
         forall evt cfg3_tgt (Htgt: lts_tgt.(step) scfg_tgt cfg_tgt evt cfg3_tgt),
           exists cfg2_src,
             rtc (lts_src.(taustep) scfg_src) cfg_src cfg2_src /\
             exists i3 cfg3_src,
               lts_src.(indexed_step) scfg_src (Indexed.mk i cfg2_src) evt (Indexed.mk i3 cfg3_src) /\
               sim i3 cfg3_src cfg3_tgt)
  .
  Hint Constructors _sim.

  Definition sim: _ -> _ -> _ -> Prop := paco3 _sim bot3.

  Lemma sim_mon: monotone3 _sim.
  Proof.
    repeat intro; destruct IN; eauto.
    eapply sim_step; eauto. intros.
    intros. exploit Hstep; eauto.
    intros [st2_src [Hst2_src [i3 [st3_src [Hst3_src Hsim]]]]].
    eexists. split; [by eauto|].
    eexists. eexists. split; [by eauto|].
    eauto.
  Qed.
  Hint Resolve sim_mon: paco.

  Lemma sim_adequacy
        (i:WF.t) (cfg_src:ltst_src.(Config)) (cfg_tgt:ltst_tgt.(Config))
        (Hsim: sim i cfg_src cfg_tgt):
    behave lts_tgt scfg_tgt cfg_tgt <1=
    behave lts_src scfg_src cfg_src.
  Proof.
    generalize dependent Hsim.
    generalize dependent cfg_tgt.
    generalize dependent cfg_src.
    generalize dependent i.
    pcofix CIH. intros.
    punfold PR. inv PR.
    generalize dependent MAT.
    generalize dependent x0.
    generalize dependent Hsim.
    generalize dependent cfg_src.
    generalize dependent i.
    apply clos_rt_rt1n in TAU. induction TAU.
    { intro i. revert x. revert i.
      refine (well_founded_ind WF.wf _ _). intros i Hi. intros.
      punfold Hsim. inv Hsim; eauto.
      - inv MAT; eauto.
        + by rewrite Herror in Htgt.
        + by rewrite Hnormal in Htgt.
        + by rewrite Hnormal in Htgt.
      - inv MAT; eauto; tac.
        + by rewrite Herror in Htgt.
        + by rewrite Hterm in Htgt.
        + exploit Hstep; eauto.
          intros [cfg2_src [Hcfg2_src [i3 [cfg3_src [Hcfg3_src Hsim]]]]]. tac.
          inv Hcfg3_src; tac.
          * exploit Hi; [by eauto|by eauto| |].
            { instantiate (1 := obs_inftau).
              clear -Hs'. punfold Hs'. inv Hs'.
              induction TAU; auto.
              eapply behmatch_inftau; eauto.
              eapply _step_normal; eauto.
            }
            intro X. punfold X. inv X.
            pfold. econstructor; [|by eauto].
            eapply rt_trans; eauto.
          * pfold. econstructor; [by eauto|].
            eapply behmatch_inftau; eauto.
            eapply _step_normal; eauto.
        + exploit Hstep; eauto.
          intros [cfg2_src [Hcfg2_src [i3 [cfg3_src [Hcfg3_src Hsim]]]]]. tac.
          inv Hcfg3_src; tac.
          pfold. econstructor; [by eauto|].
          eapply behmatch_event; eauto.
          eapply _step_normal; eauto.
    }
    { intros. punfold Hsim. inv Hsim; eauto.
      - exploit @_step_normal; eauto.
        intro X. rewrite Htgt in X. tac.
      - exploit Hstep; eauto.
        intros [cfg2_src [Hcfg2_src [i3 [cfg3_src [Hcfg3_src Hsim]]]]]. tac.
        destruct Hcfg3_src; tac.
        + exploit IHTAU; eauto.
          intro X. punfold X. inv X.
          pfold. econstructor; [|by eauto].
          eapply rt_trans; eauto.
        + exploit IHTAU; eauto.
          intro X. punfold X. inv X.
          pfold. econstructor; [|by eauto].
          eapply rt_trans; [by eauto|].
          eapply rt_trans; [|by eauto].
          by apply rt_step.
    }
  Qed.
End Simulation.
Hint Constructors _sim.
Hint Resolve sim_mon: paco.
