Require Import Equality.
Require Import Relation_Operators.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import Events.
Require Import Behaviors.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import MemoryRelation.
Require Import StateRelation.
Require Import SimulationLocal.
Require Import SimulationProgWeak.
Require Import Adequacy.
Require Import IdentityCompiler.

Import SyntaxNotations.

Program Definition foo_src :=
  FUNCTION %f"foo" WITH [%i"a"; %i"b"]
  LOCALS [%p"q"] BEGIN [
    %i"a" ::= (%i"a" MINUSii %i"b") PLUSii ((2 TIMES %i"b") MINUSii %i"b");
    %p"q" ::= CASTp %i"a";
    123 --> %p"q"
  ] END
.
Lemma foo_src_wf globals: FDecl.wf globals foo_src.
Proof. wftac. Qed.

Program Definition foo_tgt :=
  FUNCTION %f"foo" WITH [%i"a"; %i"b"]
  LOCALS [%p"q"] BEGIN [
    %p"q" ::= CASTp %i"a";
    123 --> %p"q"
  ] END
.
Lemma foo_tgt_wf globals: FDecl.wf globals foo_tgt.
Proof. wftac. Qed.

Lemma Int_mul_2 i: Int.mul 2 i = Int.add i i.
Proof.
  assert (X: (Int.repr 2) = Int.add Int.one Int.one); [done|rewrite X].
  rewrite Int.mul_add_distr_l.
  by rewrite Int.mul_commut, Int.mul_one.
Qed.

Ltac vartac :=
  rewrite ? Env.gsspec in *;
  repeat
    (match goal with
       | [H: (_, _) = (_, _) |- _] =>
         inv H
       | [H: %f ?a = %f ?a |- _] =>
         clear H
       | [H: %f ?a = %f ?b |- _] =>
         by apply ident_of_string_inject in H
       | [|- context[Var.eq ?a ?a]] =>
         destruct (Var.eq a a); try done
       | [|- context[Var.eq ?a ?b]] =>
         destruct (Var.eq a b); try done
       | [H: context[Var.eq ?a ?a] |- _] =>
         destruct (Var.eq a a); try done
       | [H: context[Var.eq ?a ?b] |- _] =>
         destruct (Var.eq a b); try done
     end).

Lemma foo_sim_local_func_aux sst_src sst_tgt (mrel_init:MemRel.t)
      (Hmrel_init_globals: mrel_genv mrel_init sst_src.(genv) sst_tgt.(genv)):
  sim_local_func_aux lts lts sst_src sst_tgt mrel_init foo_src foo_tgt.
Proof.
  unfold foo_src, foo_tgt. repeat intro.
  exploit mrel_bind_params; eauto. intros. tac.
  destruct argvals_src as [|argval1_src argvals_src]; tac.
  destruct argvals_src as [|argval2_src argvals_src]; tac.
  destruct argvals_tgt as [|argval1_tgt argvals_tgt]; tac.
  destruct argvals_tgt as [|argval2_tgt argvals_tgt]; tac.
  inv Hargvals. inv H6. inv H8. vgtac.des. inv H. rename H0 into Hlenv1.
  exists WF.elt.

  (* 0. initialize *)
  pfold. constructor; tac. intros.
  destruct cfg3_tgt as [st3_tgt|]; tac.
  Focus 2.
  { unfold sem_instr in H. tac. inv H; tac. }
  Unfocus.
  unfold sem_instr in Htgt. tac. rename mem into mem_tgt.
  exploit mrel_Env_init; eauto. intros. tac. vgtac.des.
  eexists. split.
  { apply rt_step. tac.
    instantiate (1 := Some _). tac.
    unfold sem_instr. tac.
    repeat (eapply bindPP_right; eauto; tac).
  }
  inv H. inv Henv. inv Hr1. inv Henv.

  (* 1. arithmetic *)
  destruct argval1_src as [i1_src|]; inv Hcond.
  destruct argval2_src as [i2_src|]; inv Hcond0.
  destruct argval1_tgt as [i1_tgt|]; inv Hcond1.
  destruct argval2_tgt as [i2_tgt|]; inv Hcond2.
  inv H4. inv H5.
  exists WF.elt. eexists. split.
  { apply indexed_step_step. tac.
    instantiate (1 := Some _). tac.
    unfold sem_instr. tac. vartac.
    repeat ((try eapply bindPP_right; eauto); tac).
  }
  eexists. vgtac.splits; eauto.
  { reflexivity. }
  { tac. }
  left. tac.
  rewrite Int_mul_2. rewrite Int.sub_add_l, Int.sub_idem, Int.add_zero_l.
  rewrite <- Int.sub_add_l, Int.add_commut, Int.sub_add_l.
  rewrite Int.sub_idem, Int.add_zero_l.

  (* 2. cast. *)
  destruct (Mem.denormalize i1_tgt mem_src) eqn:X; tac.
  Focus 2.
  { pfold. eapply sim_local_error; [apply rt_step|]; tac.
    - instantiate (1 := None). tac. split; tac.
      unfold sem_instr. tac. vartac. tac.
      econstructor; auto. eapply bindPP_right; eauto; vartac.
      simpl. rewrite X. constructor; auto.
    - auto.
  }
  Unfocus.
  exploit MemRel.denormalize_public; eauto. intro. des.
  pfold. apply sim_local_step; tac. intros.
  destruct cfg3_tgt as [st2_tgt|]; tac.
  Focus 2.
  { inv H2; tac.
    - inv Hr1; tac; vartac. inv Hcond.
      rewrite H1 in Hr2. inv Hr2; tac.
    - destruct Hr2; tac; vartac. inv Hcond.
      rewrite H1 in Hr1. tac.
  }
  Unfocus.
  unfold sem_instr in Htgt. tac. vartac. tac.
  rewrite H1 in Hr1. inv Hr1. destruct Hr2; tac.
  eexists. split; [by apply rt_refl|]. exists WF.elt. eexists. split.
  { apply indexed_step_step. tac.
    instantiate (1 := Some _). tac.
    unfold sem_instr. tac. vartac.
    repeat ((try eapply bindPP_right; eauto); tac).
    - rewrite X. eauto.
    - right. eauto.
  }
  tac. eexists. split; [reflexivity|]. split; tac. left.

  (* 3. store *)
  destruct (Mem.store (Value.ptr i i0) 123 mem_src) eqn:Hload_src; tac.
  Focus 2.
  { pfold. eapply sim_local_error; [apply rt_step|]; tac.
    { instantiate (1 := None). tac. split; tac.
      unfold sem_instr. tac. vartac. tac.
      repeat ((try eapply bindPP_right; eauto); tac).
      rewrite Hload_src. tac.
    }
    { tac. }
  }
  Unfocus.
  exploit H0; vartac. intros. tac.
  exploit (MemRel.store_public mrel_entry mem_src mem Hmrel_entry (Value.ptr i i0) (Value.ptr b_tgt i0)); tac.
  { by constructor. }
  { repeat instantiate (1 := 123). constructor. }
  { eauto. }
  intros. tac.
  pfold. apply sim_local_step; tac. intros.
  destruct cfg3_tgt as [st3_tgt|]; tac.
  Focus 2.
  { unfold sem_instr in H7. tac. vartac. inv H7; tac. inv Hr2; tac.
    rewrite H4 in Hr0. inv Hr0.
  }
  Unfocus.
  unfold sem_instr in Htgt. tac. vartac. tac.
  eexists. split; [apply rt_refl|]. exists WF.elt. eexists. split.
  { apply indexed_step_step. tac.
    instantiate (1 := Some _). tac.
    unfold sem_instr. tac. vartac. tac.
    repeat ((try eapply bindPP_right; eauto); tac).
    rewrite Hload_src. tac.
  }
  eexists. vgtac.splits; eauto.
  { constructor. tac. }
  left.

  (* 4. return *)
  pfold. eapply sim_local_return; eauto; tac.
  { destruct stack_src; tac. }
  { destruct stack; tac. }
  { reflexivity. }
  constructor; tac.
Qed.

Lemma foo_refinement globals funcs
      (Hwf_src: Prog.wf (Prog.mk globals (funcs ++ [foo_tgt])))
      (Hwf_tgt: Prog.wf (Prog.mk globals (funcs ++ [foo_tgt]))):
  behave_prog lts (Prog.mk globals (funcs ++ [foo_tgt])) <1=
  behave_prog lts (Prog.mk globals (funcs ++ [foo_src])).
Proof.
  apply adequacy.
  apply sim_prog.
  apply list_forall2_app.
  - unfold Prog.wf in *.
    revert Hwf_tgt. revert Hwf_src.
    generalize funcs at 1 2 7 8. intro X. induction X; constructor; tac.
    + split; tac.
      repeat intro. apply identity_sim_local_func; auto.
      * eapply FDecl.wf_mon; eauto.
      * eapply FDecl.wf_mon; eauto.
    + exploit IHX; eauto.
  - repeat constructor. by apply foo_sim_local_func_aux.
Qed.
