Require Import Equality.
Require Import Relation_Operators.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import Events.
Require Import Behaviors.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import MemoryRelation.
Require Import StateRelation.
Require Import SimulationLocal.
Require Import SimulationProgWeak.
Require Import Adequacy.
Require Import IdentityCompiler.

Import SyntaxNotations.

Program Definition foo_src :=
  FUNCTION %f"foo" WITH [%i"a"; %i"b"; %i"c1"; %i"c2"]
  LOCALS [] BEGIN [
    _::= OUTPUT (%i"a" PLUSii (%i"b" MINUSii %i"c1"));
    _::= OUTPUT (%i"a" PLUSii (%i"b" MINUSii %i"c2"))
  ] END
.
Lemma foo_src_wf globals: FDecl.wf globals foo_src.
Proof. wftac. Qed.

Program Definition foo_tgt :=
  FUNCTION %f"foo" WITH [%i"a"; %i"b"; %i"c1"; %i"c2"]
  LOCALS [%i"t"] BEGIN [
    %i"t" ::= %i"a" PLUSii %i"b";
    _::= OUTPUT (%i"t" MINUSii %i"c1");
    _::= OUTPUT (%i"t" MINUSii %i"c2")
  ] END
.
Lemma foo_tgt_wf globals: FDecl.wf globals foo_tgt.
Proof. wftac. Qed.

Ltac vartac :=
  rewrite ? Env.gsspec in *;
  repeat
    (match goal with
       | [H: (_, _) = (_, _) |- _] =>
         inv H
       | [H: %f ?a = %f ?a |- _] =>
         clear H
       | [H: %f ?a = %f ?b |- _] =>
         by apply ident_of_string_inject in H
       | [|- context[Var.eq ?a ?a]] =>
         destruct (Var.eq a a); try done
       | [|- context[Var.eq ?a ?b]] =>
         destruct (Var.eq a b); try done
       | [H: context[Var.eq ?a ?a] |- _] =>
         destruct (Var.eq a a); try done
       | [H: context[Var.eq ?a ?b] |- _] =>
         destruct (Var.eq a b); try done
     end).

Lemma foo_sim_local_func_aux sst_src sst_tgt (mrel_init:MemRel.t)
      (Hmrel_init_globals: mrel_genv mrel_init sst_src.(genv) sst_tgt.(genv)):
  sim_local_func_aux lts lts sst_src sst_tgt mrel_init foo_src foo_tgt.
Proof.
  unfold foo_src, foo_tgt. repeat intro.
  exploit mrel_bind_params; eauto. intros. simtac.
  destruct argvals_src as [|argval1_src argvals_src]; simtac.
  destruct argvals_src as [|argval2_src argvals_src]; simtac.
  destruct argvals_src as [|argval3_src argvals_src]; simtac.
  destruct argvals_src as [|argval4_src argvals_src]; simtac.
  destruct argvals_tgt as [|argval1_tgt argvals_tgt]; simtac.
  destruct argvals_tgt as [|argval2_tgt argvals_tgt]; simtac.
  destruct argvals_tgt as [|argval3_tgt argvals_tgt]; simtac.
  destruct argvals_tgt as [|argval4_tgt argvals_tgt]; simtac.
  destruct argval1_src as [i1_src|]; simtac.
  destruct argval2_src as [i2_src|]; simtac.
  destruct argval3_src as [i3_src|]; simtac.
  destruct argval4_src as [i4_src|]; simtac.
  destruct argval1_tgt as [i1_tgt|]; simtac.
  destruct argval2_tgt as [i2_tgt|]; simtac.
  destruct argval3_tgt as [i3_tgt|]; simtac.
  destruct argval4_tgt as [i4_tgt|]; simtac.
  inv Hargvals. inv H6. inv H8. inv H9. inv H10. vgtac.des. inv H. rename H0 into Hlenv1.
  inv H4. inv H5. inv H6. inv H7.
  exists WF.elt.

  (* 0. initialize *)
  pfold. apply sim_local_step; simtac. intros.
  destruct cfg3_tgt as [st3_tgt|]; simtac.
  Focus 2.
  { unfold sem_instr in H. simtac. inv H; simtac. }
  Unfocus.
  unfold sem_instr in Htgt. simtac.
  eexists. split; [apply rt_refl|].
  inv Hr1. inv Henv.
  exists (WF.from_positive 2). eexists. split.
  { apply indexed_step_step; simtac.
    instantiate (1 := Some _). simtac.
    unfold sem_instr. simtac.
    repeat (eapply bindPP_right; eauto; simtac).
    econstructor.
  }
  eexists. vgtac.splits; eauto.
  { reflexivity. }
  { simtac. }
  left.

  (* 1. arighmetic *)
  pfold. apply sim_local_step; simtac. intros.
  destruct cfg3_tgt as [st3_tgt|]; simtac.
  Focus 2.
  { unfold sem_instr in H. simtac.
    rewrite ? peq_false in *; repeat intro; simtac; vartac.
    inv H; simtac.
    inv Hr1; simtac. inv Hr0; simtac. inv Hr2; simtac.
  }
  Unfocus.
  unfold sem_instr in Htgt. simtac.
  destruct Hr3; simtac. destruct Hr0; simtac.
  rewrite ? peq_false in *; repeat intro; simtac; vartac.
  eexists. split; [apply rt_refl|].
  exists (WF.from_positive 1). eexists. split.
  { apply indexed_step_index; simtac. }
  eexists. vgtac.splits; eauto.
  { reflexivity. }
  { simtac. }
  left.

  (* 2. output 1 *)
  pfold. apply sim_local_step; simtac. intros.
  destruct cfg3_tgt as [st3_tgt|]; simtac.
  Focus 2.
  { unfold sem_instr in H. simtac.
    rewrite ? peq_false in *; repeat intro; simtac; vartac.
    inv H; simtac. inv Hr1; simtac. inv Hr0; simtac. inv Hr2; simtac.
  }
  Unfocus.
  unfold sem_instr in Htgt. simtac.
  destruct Hr0; simtac.
  rewrite ? peq_false in *; repeat intro; simtac; vartac.
  eexists. split; [apply rt_refl|]. exists WF.elt. eexists. split.
  { apply indexed_step_step. simtac.
    instantiate (1 := Some _). simtac.
    unfold sem_instr. simtac.
    rewrite ? peq_false in *; repeat intro; simtac; vartac.
    repeat (eapply bindPP_right; eauto; simtac).
    { simtac. }
    { simtac. }
    { simtac. rewrite ? Int.sub_add_opp, Int.add_assoc. eauto. }
  }
  eexists. vgtac.splits; eauto.
  { reflexivity. }
  { simtac. }
  left.

  (* 3. output 2 *)
  pfold. apply sim_local_step; simtac. intros.
  destruct cfg3_tgt as [st3_tgt|]; simtac.
  Focus 2.
  { unfold sem_instr in H. simtac.
    rewrite ? peq_false in *; repeat intro; simtac; vartac.
    inv H; simtac. inv Hr1; simtac. inv Hr0; simtac. inv Hr2; simtac.
  }
  Unfocus.
  unfold sem_instr in Htgt. simtac.
  destruct Hr0; simtac.
  rewrite ? peq_false in *; repeat intro; simtac; vartac.
  eexists. split; [apply rt_refl|]. exists WF.elt. eexists. split.
  { apply indexed_step_step. simtac.
    instantiate (1 := Some _). simtac.
    unfold sem_instr. simtac.
    rewrite ? peq_false in *; repeat intro; simtac; vartac.
    repeat (eapply bindPP_right; eauto; simtac).
    { simtac. }
    { simtac. }
    { simtac. rewrite ? Int.sub_add_opp, Int.add_assoc. eauto. }
  }
  eexists. vgtac.splits; eauto.
  { reflexivity. }
  { simtac. }
  left.

  (* 4. return *)
  pfold. eapply sim_local_return; eauto; simtac.
  { destruct stack_src; simtac. }
  { destruct stack_tgt; simtac. }
  { reflexivity. }
  { reflexivity. }
Qed.

Lemma foo_refinement globals funcs
      (Hwf_src: Prog.wf (Prog.mk globals (funcs ++ [foo_tgt])))
      (Hwf_tgt: Prog.wf (Prog.mk globals (funcs ++ [foo_tgt]))):
  behave_prog lts (Prog.mk globals (funcs ++ [foo_tgt])) <1=
  behave_prog lts (Prog.mk globals (funcs ++ [foo_src])).
Proof.
  apply adequacy.
  apply sim_prog.
  apply list_forall2_app.
  - unfold Prog.wf in *.
    revert Hwf_tgt. revert Hwf_src.
    generalize funcs at 1 2 7 8. intro X. induction X; constructor; tac.
    + split; tac.
      repeat intro. apply identity_sim_local_func; auto.
      * eapply FDecl.wf_mon; eauto.
      * eapply FDecl.wf_mon; eauto.
    + exploit IHX; eauto.
  - repeat constructor. by apply foo_sim_local_func_aux.
Qed.
