Require Import Equality.
Require Import Relation_Operators.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import MemoryC.
Require Import Events.
Require Import Behaviors.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import SemanticsC.
Require Import MemoryRelationC.
Require Import StateRelationC.
Require Import SimulationLocalC.
Require Import SimulationProgWeakC.
Require Import AdequacyC.
Require Import DeadCastEliminationCompiler.

Import SyntaxNotations.

Program Definition foo :=
  FUNCTION %f"foo" WITH [%p"p"; %i"n"]
  LOCALS [%p"q"; %i"a"; %i"r"] BEGIN [
    %p"q" ::= MALLOC %i"n";
    %i"a" ::= CASTi %p"p";
    %i"r" ::= %i"a" TIMES 123
  ] END
.
Lemma foo_wf globals: FDecl.wf globals foo.
Proof. wftac. Qed.

Program Definition gee_src :=
  FUNCTION %f"gee" WITH [%p"p"; %i"n"]
  LOCALS [] BEGIN [
    CALL %f"foo" WITH [%p"p"; %i"n"];
    CALL %f"bar" WITH []
  ] END
.
Lemma gee_src_wf globals: FDecl.wf globals gee_src.
Proof. wftac. Qed.

Program Definition gee_tgt :=
  FUNCTION %f"gee" WITH [%p"p"; %i"n"]
  LOCALS [] BEGIN [
    CALL %f"bar" WITH []
  ] END
.
Lemma gee_tgt_wf globals: FDecl.wf globals gee_tgt.
Proof. wftac. Qed.

Ltac vartac :=
  rewrite ? Env.gsspec in *;
  repeat
    (match goal with
       | [H: (_, _) = (_, _) |- _] =>
         inv H
       | [H: %f ?a = %f ?a |- _] =>
         clear H
       | [H: %f ?a = %f ?b |- _] =>
         by apply ident_of_string_inject in H
       | [|- context[Var.eq ?a ?a]] =>
         destruct (Var.eq a a); try done
       | [|- context[Var.eq ?a ?b]] =>
         destruct (Var.eq a b); try done
       | [H: context[Var.eq ?a ?a] |- _] =>
         destruct (Var.eq a a); try done
       | [H: context[Var.eq ?a ?b] |- _] =>
         destruct (Var.eq a b); try done
     end).

Lemma gee_sim_local_func_aux sst_src sst_tgt (mrel_init:MemRelC.t)
      (Hfoo_src: PTree.get (%f"foo") sst_src.(Semantics.fenv) = Some foo)
      (Hmrel_init_globals: mrel_genv mrel_init sst_src.(Semantics.genv) sst_tgt.(genv)):
  sim_local_func_aux lts ltsC sst_src sst_tgt mrel_init gee_src gee_tgt.
Proof.
  unfold gee_src, gee_tgt. repeat intro.
  exploit mrel_bind_params; eauto. intros. simtac.
  destruct argvals_src as [|argval1_src argvals_src]; simtac.
  destruct argvals_src as [|argval2_src argvals_src]; simtac.
  destruct argvals_tgt as [|i1_tgt argvals_tgt]; simtac.
  destruct argvals_tgt as [|i2_tgt argvals_tgt]; simtac.
  inv Hargvals. inv H6. inv H8. vgtac.des. inv H. rename H0 into Hlenv1.
  exists WF.elt.

  (* 0. initialize *)
  destruct argval1_src as [i1_src|]; inv Hcond.
  destruct argval2_src as [i2_src|]; inv Hcond0.
  pfold. apply sim_local_step; simtac. intros.
  destruct cfg3_tgt as [st3_tgt|]; simtac.
  Focus 2.
  { inv H. simtac. inv Hr2. }
  Unfocus.
  inv Htgt. simtac. inv Hr1. inv Hr2.
  eexists. split; [apply rt_refl|].
  exists (WF.from_positive 1004). eexists. split.
  { apply indexed_step_step; simtac.
    instantiate (1 := Some _). simtac.
    unfold sem_instr. simtac.
    repeat (eapply bindPP_right; eauto; simtac).
    repeat econstructor.
  }
  eexists. vgtac.splits; eauto.
  { reflexivity. }
  { simtac. }
  left.

  exploit MemRelC.alloc_src; eauto.
  { unfold Mem.alloc. simtac. }
  intro. simtac.
  inv H4. inv Hmrel_entry.
  exploit MemRelC.ptrintcast_public.
  { eapply Mem.alloc_wf; eauto. by unfold Mem.alloc. }
  { eauto. }
  Focus 2.
  { eapply H0. }
  Unfocus.
  { eauto. }
  { econstructor; eauto. }
  intro. simtac.
  instantiate (3 := i2_src) in H2. instantiate (3 := o) in H2. inv H1.

  pfold. eapply sim_local_call; simtac.
  { eapply rt_trans; [apply rt_step|]; simtac.
    { (* 1. call *)
      instantiate (1 := Some _). simtac.
      unfold sem_instr. simtac.
      rewrite Hfoo_src. simtac.
      repeat (eapply bindPP_right; eauto; simtac).
    }
    eapply rt_trans; [apply rt_step|]; simtac.
    { (* 2. initialize *)
      instantiate (1 := Some _). simtac.
      unfold sem_instr. simtac.
      repeat (eapply bindPP_right; eauto; simtac).
      repeat constructor.
    }
    eapply rt_trans; [apply rt_step|]; simtac.
    { (* 2. malloc *)
      instantiate (1 := Some _). simtac.
      unfold sem_instr. simtac.
      rewrite ? peq_false in *; repeat intro; simtac; vartac.
      repeat (eapply bindPP_right; eauto; simtac).
      right. simtac.
    }
    eapply rt_trans; [apply rt_step|]; simtac.
    { (* 3. cast *)
      instantiate (1 := Some _). simtac.
      unfold sem_instr. simtac.
      rewrite ? peq_false in *; repeat intro; simtac; vartac.
      destruct (Mem.normalize b o (PTree.set (Mem.nextb (MemRelC.src mrel_entry)) (Mem.Block.init i2_src) (MemRelC.src mrel_entry))) eqn:X; eauto.
      - repeat (try (eapply bindPP_right; eauto); simtac).
        + rewrite X. eauto.
        + eauto.
        + right. auto.
      - repeat (try (eapply bindPP_right; eauto); simtac).
        + rewrite X. repeat (eapply bindPP_right; eauto).
        + right. auto.
    }
    eapply rt_trans; [apply rt_step|]; simtac.
    { (* 4. arithmetic *)
      instantiate (1 := Some _). simtac.
      unfold sem_instr. simtac.
      rewrite ? peq_false in *; repeat intro; simtac; vartac.
      repeat (try (eapply bindPP_right; eauto); simtac).
    }
    eapply rt_trans; [apply rt_step|]; simtac.
    { (* 5. return *)
      instantiate (1 := Some _). simtac.
    }
    apply rt_refl.
  }
  { auto. }
  { eauto. }
  { etransitivity; eauto.
    apply MemRelC.le_public_le. auto.
  }
  { reflexivity. }
  { auto. }
  { reflexivity. }
  { constructor. }
  simpl. intros.
  destruct st3_src as [ctx5_src stack5_src mem5_src].
  destruct st3_tgt as [ctx5_tgt stack5_tgt mem5_tgt]. tac.
  simpl in *. subst.

  (* 6. exit *)
  intros. exists WF.elt. left.
  pfold. eapply sim_local_return; eauto.
  { reflexivity. }
  rewrite <- Hmrel3_le, <- H4.
  constructor; simtac.
  { repeat intro. simtac.
    destruct (peq i (Mem.nextb (MemRelC.src mrel_entry))); simtac.
    exploit Mem._nextb; eauto. xomega.
  }
Qed.

Lemma gee_foo_refinement globals funcs
      (Hwf_src: Prog.wf (Prog.mk globals ((map remove_dead_cast_func funcs) ++ [gee_tgt; foo])))
      (Hwf_tgt: Prog.wf (Prog.mk globals (funcs ++ [gee_tgt; foo]))):
  behave_prog ltsC (Prog.mk globals ((map remove_dead_cast_func funcs) ++ [gee_tgt; foo])) <1=
  behave_prog lts (Prog.mk globals (funcs ++ [gee_src; foo])).
Proof.
  apply adequacyC.
  apply sim_prog.
  apply list_forall2_app.
  { unfold Prog.wf in *. simtac.
    revert Hwf_tgt. revert Hwf_src.
    generalize funcs at 1 2 5 6. intro X. induction X; constructor; simtac.
    + split; simtac.
      repeat intro. constructor.
      { simpl. induction (a.(FDecl.params)); constructor; simtac. }
      apply remove_dead_cast_sim_local_func_aux; auto.
      eapply FDecl.wf_mon; eauto.
      eapply FDecl.wf_mon; eauto.
    + exploit IHX; eauto.
  }
  repeat constructor.
  - apply gee_sim_local_func_aux; auto.
    rewrite Hfenv_src. rewrite <- FDecl_find.
    rewrite rev_app_distr. unfold FDecl.byid. simtac.
    destruct (peq (%f "foo") (%f "foo")); simtac.
  - apply remove_dead_cast_sim_local_func_aux; eauto using foo_wf.
Qed.
