Require Import Equality.
Require Import Relation_Operators.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import Events.
Require Import Behaviors.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import MemoryRelation.
Require Import StateRelation.
Require Import SimulationLocal.
Require Import SimulationProgWeak.
Require Import Adequacy.
Require Import IdentityCompiler.

Import SyntaxNotations.

Program Definition foo :=
  FUNCTION %f"foo" WITH [%i"a"]
  LOCALS [] BEGIN [
    %i"a" ::= %i"a" AND 123
  ] END
.
Lemma foo_wf globals: FDecl.wf globals foo.
Proof. wftac. Qed.

Program Definition gee_src :=
  FUNCTION %f"gee" WITH [%p"p"]
  LOCALS [%i"a"] BEGIN [
    %i"a" ::= CASTi %p"p";
    CALL %f"foo" WITH [%i"a"];
    CALL %f"bar" WITH []
  ] END
.
Lemma gee_src_wf globals: FDecl.wf globals gee_src.
Proof. wftac. Qed.

Program Definition gee_tgt :=
  FUNCTION %f"gee" WITH [%p"p"]
  LOCALS [%i"a"] BEGIN [
    %i"a" ::= CASTi %p"p";
    CALL %f"bar" WITH []
  ] END
.
Lemma gee_tgt_wf globals: FDecl.wf globals gee_tgt.
Proof. wftac. Qed.

Ltac vartac :=
  rewrite ? Env.gsspec in *;
  repeat
    (match goal with
       | [H: (_, _) = (_, _) |- _] =>
         inv H
       | [H: %f ?a = %f ?a |- _] =>
         clear H
       | [H: %f ?a = %f ?b |- _] =>
         by apply ident_of_string_inject in H
       | [|- context[Var.eq ?a ?a]] =>
         destruct (Var.eq a a); try done
       | [|- context[Var.eq ?a ?b]] =>
         destruct (Var.eq a b); try done
       | [H: context[Var.eq ?a ?a] |- _] =>
         destruct (Var.eq a a); try done
       | [H: context[Var.eq ?a ?b] |- _] =>
         destruct (Var.eq a b); try done
     end).

Lemma gee_sim_local_func_aux sst_src sst_tgt (mrel_init:MemRel.t)
      (Hfoo_src: PTree.get (%f"foo") sst_src.(fenv) = Some foo)
      (Hmrel_init_globals: mrel_genv mrel_init sst_src.(genv) sst_tgt.(genv)):
  sim_local_func_aux lts lts sst_src sst_tgt mrel_init gee_src gee_tgt.
Proof.
  unfold gee_src, gee_tgt. repeat intro.
  exploit mrel_bind_params; eauto. intros. simtac.
  destruct argvals_src as [|argval1_src argvals_src]; simtac.
  destruct argvals_tgt as [|argval1_tgt argvals_tgt]; simtac.
  inv Hargvals. inv H6. vgtac.des. inv H. rename H0 into Hlenv1.
  exists WF.elt.

  (* 0. initialize *)
  pfold. apply sim_local_step; simtac. intros.
  destruct cfg3_tgt as [st3_tgt|]; simtac.
  Focus 2.
  { unfold sem_instr in H. simtac. inv H; simtac. }
  Unfocus.
  unfold sem_instr in Htgt. simtac.
  eexists. split; [apply rt_refl|].
  inv Hr1. inv Henv.
  exists (WF.from_positive 2). eexists. split.
  { apply indexed_step_step; simtac.
    instantiate (1 := Some _). simtac.
    unfold sem_instr. simtac.
    repeat (eapply bindPP_right; eauto; simtac).
    repeat econstructor.
  }
  eexists. vgtac.splits; eauto.
  { reflexivity. }
  { simtac. }
  left.

  (* 1. cast *)
  destruct argval1_src as [|b1_src o1_src]; inv Hcond.
  destruct argval1_tgt as [|b1_tgt o1_tgt]; inv Hcond0.
  inv H4.
  pfold. apply sim_local_step; simtac. intros.
  destruct cfg3_tgt as [st3_tgt|]; simtac.
  Focus 2.
  { unfold sem_instr in H. simtac.
    inv H; tac.
    - inv Hr1; tac.
      destruct (Mem.normalize b1_tgt o1_tgt mem_tgt); tac.
      + by inv Hr2.
      + inv Hr2; [|done]. inv Hr1; [done|]. by inv Hr2.
    - destruct Hr2; tac.
      destruct t; [done|].
      destruct (Mem.normalize b1_tgt o1_tgt mem_tgt); tac.
  }
  Unfocus.
  unfold sem_instr in Htgt. tac. destruct Hr2; tac.
  vartac. inv Hcond.
  exploit MemRel.ptrintcast_public; eauto.
  { by inv Hmrel_entry. }
  { by inv Hmrel_entry. }
  intro. des.

  eexists. split; [apply rt_refl|].
  exists (WF.from_positive 1004). eexists. split.
  { apply indexed_step_step; simtac.
    instantiate (1 := Some _). simtac.
    unfold sem_instr. simtac.
    repeat (eapply bindPP_right; eauto; simtac).
    right. auto.
  }
  subst. eexists. vgtac.splits.
  { inv H3. eauto. }
  { simtac. }
  subst.
  left.

  (* 2. call *)
  pfold. eapply sim_local_call; simtac.
  { eapply rt_trans; [apply rt_step|]; simtac.
    { instantiate (1 := Some _). simtac.
      unfold sem_instr. simtac.
      rewrite Hfoo_src. simtac.
      repeat (eapply bindPP_right; eauto; simtac).
    }
    eapply rt_trans; [apply rt_step|]; simtac.
    { instantiate (1 := Some _). simtac.
      unfold sem_instr. simtac.
      repeat (eapply bindPP_right; eauto; simtac).
      econstructor.
    }
    eapply rt_trans; [apply rt_step|]; simtac.
    { instantiate (1 := Some _). simtac.
      unfold sem_instr. simtac.
      repeat (eapply bindPP_right; eauto; simtac).
      right. split; eauto.
    }
    eapply rt_trans; [apply rt_step|]; simtac.
    { instantiate (1 := Some _). simtac. }
    apply rt_refl.
  }
  { simtac. }
  { eauto. }
  { reflexivity. }
  { simtac. }
  { simtac. }
  { simtac. }
  { constructor. }
  intros. exists WF.elt. left.

  (* 3. return *)
  destruct st3_src as [ctx5_src stack5_src mem5_src].
  destruct st3_tgt as [ctx5_tgt stack5_tgt mem5_tgt]. tac.

  pfold. eapply sim_local_return; eauto.
  { reflexivity. }
  by rewrite <- Hmrel3_le.
Qed.

Lemma gee_foo_refinement globals funcs
      (Hwf_src: Prog.wf (Prog.mk globals (funcs ++ [gee_tgt; foo])))
      (Hwf_tgt: Prog.wf (Prog.mk globals (funcs ++ [gee_tgt; foo]))):
  behave_prog lts (Prog.mk globals (funcs ++ [gee_tgt; foo])) <1=
  behave_prog lts (Prog.mk globals (funcs ++ [gee_src; foo])).
Proof.
  apply adequacy.
  apply sim_prog.
  apply list_forall2_app.
  { unfold Prog.wf in *. tac.
    revert Hwf_tgt. revert Hwf_src.
    generalize funcs at 1 2 5 6. intro X. induction X; constructor; tac.
    + split; tac.
      repeat intro. apply identity_sim_local_func; auto.
      * eapply FDecl.wf_mon; eauto.
      * eapply FDecl.wf_mon; eauto.
    + exploit IHX; eauto.
  }
  repeat constructor.
  - apply gee_sim_local_func_aux; auto.
    rewrite Hfenv_src. rewrite <- FDecl_find.
    rewrite rev_app_distr. unfold FDecl.byid. tac.
    destruct (peq (%f "foo") (%f "foo")); tac.
  - eapply identity_sim_local_func_aux; eauto using foo_wf.
Qed.
