Require Import Equality.
Require Import Relation_Operators.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import Events.
Require Import Behaviors.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import MemoryRelation.
Require Import StateRelation.
Require Import SimulationLocal.

Import SyntaxNotations.

Lemma identity_sim_local_func_aux
      sst_src sst_tgt (mrel_init:MemRel.t) foo
      (Hfoo_wf_src: FDecl.wf (PTree.keys sst_src.(genv)) foo)
      (Hfoo_wf_tgt: FDecl.wf (PTree.keys sst_tgt.(genv)) foo)
      (Hmrel_init_globals: mrel_genv mrel_init sst_src.(genv) sst_tgt.(genv)):
  sim_local_func_aux lts lts sst_src sst_tgt mrel_init
                 foo foo.
Proof.
  repeat intro.
  exploit mrel_bind_params; eauto.
  intro. vgtac.des.
  rewrite Henv_tgt in Hlenv_tgt. inv Hlenv_tgt. simtac.
  exists WF.elt.
  cut (forall
          lenv_src lenv_tgt (mrel:MemRel.t) mem_src mem_tgt istrs
          (Hwf_src: Context_wf sst_src (mkContext lenv_src istrs))
          (Hwf_tgt: Context_wf sst_tgt (mkContext lenv_tgt istrs))
          (Hmrel_le_public: MemRel.le_public mrel_entry mrel)
          (Hmrel: mrel mem_src mem_tgt)
          (Hlenv: mrel_env mrel mem_src mem_tgt lenv_src lenv_tgt),
       sim_local lts lts sst_src sst_tgt stack_src stack_tgt mrel_entry mrel WF.elt
                 (Some (mkState (mkContext lenv_src istrs) stack_src mem_src))
                 (Some (mkState (mkContext lenv_tgt istrs) stack_tgt mem_tgt))).
  { intro X. apply X; simtac.
    - eapply Context_wf_before_initialize; simtac.
      + eapply bind_params_wf; eauto.
      + apply Forall_forall. intros i Hi.
        eapply for_all_spec in Hfoo_wf_src; eauto.
        eapply Instr.wf_mon; eauto; tac.
        intros v Hv. rewrite in_app in *. destruct Hv as [Hv|Hv].
        * left. exploit (bind_params_keys2 (FDecl.params foo) argvals_src); eauto.
          rewrite Env.gkspec. by destruct (Env.get v lenv_src).
        * by right.
    - eapply Context_wf_before_initialize; simtac.
      + eapply bind_params_wf; eauto.
      + apply Forall_forall. intros i Hi.
        eapply for_all_spec in Hfoo_wf_tgt; eauto.
        eapply Instr.wf_mon; eauto; tac.
        intros v Hv. rewrite in_app in *. destruct Hv as [Hv|Hv].
        * left. exploit (bind_params_keys2 (FDecl.params foo) argvals_tgt); eauto.
          rewrite Env.gkspec. by destruct (Env.get v env_tgt).
        * by right.
    - reflexivity.
  }
  clear lenv_src env_tgt Hlenv_src Henv_tgt Henv.
  clear mem_src mem_tgt Hmrel_entry Hargvals.
  clear argvals_src argvals_tgt.
  pcofix CIH. intros.
  destruct istrs as [|istr istrs].
  { pfold. eapply sim_local_return; eauto; simtac.
    - destruct stack_src; simtac.
    - destruct stack_tgt; simtac.
    - reflexivity.
  }
  destruct (classic
              (sem_instr sst_src
                         (mkState (mkContext lenv_src (istr::istrs)) stack_src mem_src)
                         None)).
  { pfold. eapply sim_local_error.
    apply rt_step. simtac.
    - instantiate (1 := None). simtac.
    - simtac.
  }
  destruct (Instr.is_call istr) eqn:Hcall; simtac.
  { destruct istr; simtac.
    destruct (sem_bexp_list (genv sst_src) lenv_src args) eqn: Hargvals_src; simtac.
    Focus 2.
    { contradict H. unfold sem_instr. simtac.
      destruct ((fenv sst_src) ! fid); simtac.
      apply bindPP_left_None; simtac.
    }
    Unfocus.
    inv Hwf_src; simtac. inv Hinstrs; simtac.
    inv Hwf_tgt; simtac. inv Hinstrs; simtac.
    exploit mrel_bexp_list; simtac.
    { instantiate (3 := mrel). eapply mrel_genv_le; eauto.
      inv Hmrel_init. inv Hmrel_le_public. inv Hequiv. auto.
    }
    { eauto. }
    { eauto. }
    { auto. }
    { auto. }
    { eauto. }
    { eauto. }
    intros. vgtac.des. pfold. eapply sim_local_call; eauto; simtac.
    - reflexivity.
    - intros. exists WF.elt. right.
      exploit (CIH lenv_src lenv_tgt mrel3 st3_src.(mem) st3_tgt.(mem)); eauto.
      + instantiate (1 := istrs).
        constructor; simtac.
      + constructor; simtac.
      + by rewrite Hmrel_le_public.
      + eapply (mrel_env_le lenv_src lenv_tgt mem_src); eauto.
        by inv Hmrel3_le.
      + destruct st3_src as [ctx3_src stack3_src mem3_src]. simtac.
        destruct st3_tgt as [ctx3_tgt stack3_tgt mem3_tgt]. simtac.
  }
  pfold. apply sim_local_step; simtac. intros.
  destruct cfg3_tgt as [st3_tgt|]; simtac.
  Focus 2.
  { contradict H.
    eapply mrel_instr_None; try apply H0; simtac; auto;
      [|by eauto|by eauto].
    eapply mrel_genv_le; eauto.
    inv Hmrel_init. inv Hmrel_le_public. inv Hequiv. auto.
  }
  Unfocus.
  exploit
  (mrel_instr
     istr istrs
     (mkState (mkContext lenv_src (istr::istrs)) stack_src mem_src)
     (mkState (mkContext lenv_tgt (istr::istrs)) stack_tgt mem_tgt)
     mrel);
    simtac; eauto.
  { eapply mrel_genv_le; [by eauto|].
    inv Hmrel_le_public. inv Hequiv.
    inv Hmrel_init. auto.
  }
  intro. vgtac.des.
  { by contradict H. }
  eexists. split; [apply rt_refl|].
  exists WF.elt. eexists. split.
  - apply indexed_step_step. simtac.
    instantiate (1 := Some _). simtac. eauto.
  - eexists. vgtac.splits.
    + apply MemRel.le_public_le. eauto.
    + simtac.
    + right.
      destruct st2_src as [ctx3_src stack3_src mem3_src]. simtac.
      destruct st3_tgt as [ctx3_tgt stack3_tgt mem3_tgt]. simtac.
      exploit sem_instr_not_call_prop; eauto; simtac. intro. simtac.
      exploit sem_instr_wf; eauto; simtac. intro Hctx3_src_wf.
      move Htgt at bottom.
      exploit sem_instr_not_call_prop; eauto; simtac. intro. simtac.
      exploit sem_instr_wf; eauto; simtac. intro Hctx3_tgt_wf.
      exploit
        (CIH ctx3_src.(lenv) ctx3_tgt.(lenv)
                                        mrel2 mem3_src mem3_tgt); eauto.
      { instantiate (1 := ctx3_src.(instrs)).
        by destruct ctx3_src.
      }
      { rewrite Histrs. by destruct ctx3_tgt. }
      { by rewrite Hmrel_le_public. }
      { destruct ctx3_src, ctx3_tgt. simtac. }
Qed.

Lemma identity_sim_local_func
      sst_src sst_tgt (mrel_init:MemRel.t) foo
      (Hfoo_wf_src: FDecl.wf (PTree.keys sst_src.(genv)) foo)
      (Hfoo_wf_tgt: FDecl.wf (PTree.keys sst_tgt.(genv)) foo)
      (Hmrel_init_globals: mrel_genv mrel_init sst_src.(genv) sst_tgt.(genv)):
  sim_local_func lts lts sst_src sst_tgt mrel_init
                 foo foo.
Proof.
  constructor.
  - induction (FDecl.params foo); constructor; tac.
  - by apply identity_sim_local_func_aux.
Qed.
