Require Import Equality.
Require Import Relation_Operators.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import Events.
Require Import Behaviors.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import MemoryRelation.
Require Import StateRelation.
Require Import SimulationLocal.
Require Import SimulationProgWeak.
Require Import Adequacy.
Require Import IdentityCompiler.

Import SyntaxNotations.

Program Definition foo_src :=
  FUNCTION %f"foo" WITH [%p"p"]
  LOCALS [%p"q"; %i"a"] BEGIN [
    %p"q" ::= MALLOC 1;
    123 --> %p"q";
    CALL %f"bar" WITH [%p"p"];
    %i"a" <-- %p"q";
    %i"a" --> %p"p"
  ] END
.
Lemma foo_src_wf globals: FDecl.wf globals foo_src.
Proof. wftac. Qed.

Program Definition foo_tgt :=
  FUNCTION %f"foo" WITH [%p"p"]
  LOCALS [] BEGIN [
    CALL %f"bar" WITH [%p"p"];
    123 --> %p"p"
  ] END
.
Lemma foo_tgt_wf globals: FDecl.wf globals foo_tgt.
Proof. wftac. Qed.

Ltac tac :=
  repeat (simtac;
          try match goal with
                | [H: bindPP _ _ None |- _] =>
                  inv H
              end).

Lemma range_01: Range.contained 0 (Range.mk 0 1).
Proof. tac. Qed.

Ltac vartac :=
  rewrite ? Env.gsspec in *;
  repeat
    (match goal with
       | [H: (_, _) = (_, _) |- _] =>
         inv H
       | [H: %f ?a = %f ?a |- _] =>
         clear H
       | [H: %f ?a = %f ?b |- _] =>
         by apply ident_of_string_inject in H
       | [|- context[Var.eq ?a ?a]] =>
         destruct (Var.eq a a); try done
       | [|- context[Var.eq ?a ?b]] =>
         destruct (Var.eq a b); try done
       | [H: context[Var.eq ?a ?a] |- _] =>
         destruct (Var.eq a a); try done
       | [H: context[Var.eq ?a ?b] |- _] =>
         destruct (Var.eq a b); try done
     end).

Lemma foo_sim_local_func_aux sst_src sst_tgt (mrel_init:MemRel.t)
      (Hmrel_init_globals: mrel_genv mrel_init sst_src.(genv) sst_tgt.(genv)):
  sim_local_func_aux lts lts sst_src sst_tgt mrel_init foo_src foo_tgt.
Proof.
  unfold foo_src, foo_tgt. repeat intro.
  exploit mrel_bind_params; eauto. intros. tac.
  destruct argvals_src as [|argval1_src argvals_src]; tac.
  destruct argvals_tgt as [|argval1_tgt argvals_tgt]; tac.
  inv Hargvals. inv H6. vgtac.des. inv H. rename H0 into Hlenv1.
  exists WF.elt.

  (* 0. initialize *)
  pfold. constructor; tac. intros.
  unfold sem_instr in Htgt. destruct cfg3_tgt as [st3_tgt|]; tac.
  rename v1 into lenv_tgt.
  eexists. split; [by apply rt_refl|]. exists WF.elt.
  exploit Env.init_nonempty. intros [lenv_src Hlenv_src]. tac.
  eexists. split.
  { apply indexed_step_step. tac.
    instantiate (1 := Some _). tac.
    unfold sem_instr. tac.
    repeat (eapply bindPP_right; eauto; tac).
  }

  eexists. split; [reflexivity|]. split; tac.
  inv Hr1. inv Hlenv_src. inv Henv. inv Henv0. left.

  (* 1. malloc *)
  exploit MemRel.alloc_src; eauto; [by unfold Mem.alloc|].
  instantiate (1 := 1). intros [Hmrel12 Hmrel2].

  (* 2. store *)
  exploit MemRel.store_src; eauto; tac.
  { instantiate (3 := Value.ptr mem_src.(Mem.nextb) 0).
    unfold Mem.val2ptr. tac.
  }
  { rewrite peq_true. eauto. }
  { unfold Mem.store. tac.
    unfold Mem.val2ptr. tac.
    rewrite range_01.
    unfold Mem.store_aux. tac.
  }
  intros [src3_private [Hsrc3_private [Hmrel23 Hmrel3]]].
  unfold Mem.store_aux in Hsrc3_private. tac.
  unfold Mem.Block.store in Hcond1. tac.
  instantiate (1 := 123) in Hmrel3.
  rewrite Hmrel23 in Hmrel12. clear Hmrel23. rename Hmrel12 into Hmrel13. clear Hmrel2.

  (* 3. call *)
  pfold. eapply sim_local_call; tac.
  { eapply rt_trans.
    { apply rt_step. tac.
      instantiate (1 := Some _). tac.
      unfold sem_instr. tac.
      repeat (try (eapply bindPP_right; eauto); tac).
    }
    { apply rt_step. tac.
      unfold sem_instr. tac.
      repeat (try (eapply bindPP_right; eauto); tac).
      unfold Mem.store, Mem.val2ptr. tac. rewrite range_01.
      unfold Mem.store_aux. tac.
    }
  }
  { tac. }
  { tac. eauto. }
  { tac. }
  { tac. }
  { tac. }
  { by rewrite Hcond0. }
  { tac.
    rewrite ? peq_false in *; repeat intro; tac; vartac.
    rewrite Hcond. tac.
  }
  { tac. repeat constructor.
    exploit MemRel.incr_rel_value; eauto.
  }
  destruct st3_src as [ctx3_src stack3_src mem3_src].
  destruct st3_tgt as [ctx3_tgt stack3_tgt mem3_tgt]. tac.
  intros. exists WF.elt. tac. left.

  (* 4. load *)
  destruct (Mem.load (Value.ptr (Mem.nextb mem_src) Int.zero) mem3_src) eqn:Hload4_src.
  Focus 2.
  { pfold. eapply sim_local_error.
    { apply rt_step. tac.
      instantiate (1 := None). tac. split; tac.
      unfold sem_instr. tac.
      repeat (try (eapply bindPP_right; eauto); tac).
      by rewrite Hload4_src.
    }
    { tac. }
  }
  Unfocus.

  assert (Y: mrel3.(MemRel.src_private) ! (Mem.nextb mem_src) = Some (Mem.Block.mk true true 1 None (ZMap.set (Int.unsigned 0) (Value.int 123) (ZMap.init (Value.int Int.zero))))).
  { inv Hmrel3_le. tac. apply Hsrc_private. tac. }
  assert (Z: mem3_src ! (Mem.nextb mem_src) = Some (Mem.Block.mk true true 1 None (ZMap.set (Int.unsigned 0) (Value.int 123) (ZMap.init (Value.int Int.zero))))).
  { erewrite get_src; eauto. }
  erewrite MemRel.load_src in Hload4_src; eauto.
  Focus 2.
  { unfold Mem.val2ptr. tac. rewrite Z. tac. }
  Unfocus.
  unfold Mem.load_aux in Hload4_src. tac. unfold Mem.Block.load in H0. tac.

  (* 5. store *)
  destruct (Mem.store argval1_src 123 mem3_src) eqn:Hstore5_src.
  Focus 2.
  { pfold. eapply sim_local_error.
    { eapply rt_trans; [apply rt_step|].
      { tac. instantiate (1 := Some _). tac.
        unfold sem_instr. tac.
        repeat (try (eapply bindPP_right; eauto); tac).
        unfold Mem.load, Mem.val2ptr. tac. rewrite Z. tac.
        assert (W: Int.zero = 0); tac. rewrite W in *.
        rewrite Hcond3. unfold Mem.load_aux. rewrite Z. unfold Mem.Block.load. tac.
        rewrite W, Hcond3. tac.
      }
      eapply rt_trans; [apply rt_step|].
      { tac. instantiate (1 := None). tac.
        unfold sem_instr. tac. split; tac.
        rewrite ? peq_false in *; repeat intro; tac; vartac.
        repeat (try (eapply bindPP_right; eauto); tac).
        { by rewrite Hcond. }
        { rewrite Hstore5_src. tac. }
      }
      apply rt_refl.
    }
    { tac. }
  }
  Unfocus.

  exploit MemRel.store_public; eauto.
  { generalize H4. intro Y. eapply MemRel.incr_rel_value in Y; eauto.
    rewrite Hmrel13. by apply MemRel.le_public_le.
  }
  { instantiate (1 := 123). constructor. }
  intro. tac.

  pfold. apply sim_local_step; tac. intros.
  destruct cfg3_tgt as [st3_tgt|]; tac.
  Focus 2.
  { unfold sem_instr in H2. tac.
    { by rewrite Hcond0 in Hr1. }
    rewrite H in Hr0. inv Hr0.
  }
  Unfocus.
  unfold sem_instr in Htgt. tac.
  eexists. split.
  { apply rt_step. tac.
    instantiate (1 := Some _). tac.
    unfold sem_instr. tac.
    repeat (try (eapply bindPP_right; eauto); tac).
    unfold Mem.load, Mem.val2ptr. tac. rewrite Z. tac.
    assert (W: Int.zero = 0); tac. rewrite W in *. rewrite Hcond3.
    unfold Mem.load_aux. rewrite Z. unfold Mem.Block.load. tac.
    rewrite W, Hcond3. tac.
  }
  exists WF.elt. eexists. split.
  { apply indexed_step_step. tac.
    instantiate (1 := Some _). tac.
    unfold sem_instr. tac.
    rewrite ? peq_false in *; repeat intro; tac; vartac.
    repeat (try (eapply bindPP_right; eauto); tac).
    { by rewrite Hcond. }
    rewrite Hstore5_src. tac.
  }
  eexists. vgtac.splits; eauto.
  { tac. }
  left.

  (* 6. return *)
  pfold. eapply sim_local_return; eauto; tac.
  { by destruct stack_src. }
  { by destruct stack_tgt. }
  { reflexivity. }
  constructor; tac.
  { rewrite Hmrel13, <- H0.
    apply MemRel.le_public_le. auto.
  }
  { repeat intro.
    inv Hmrel3_le. tac. apply Hsrc_private. tac.
    destruct (peq i mem_src.(Mem.nextb)); tac.
    inv Hmrel_entry. apply Hsrc_private_subset in H2.
    inv Hsrc_wf. exploit _nextb; eauto. intro W. by apply Plt_ne in W.
  }
  { repeat intro.
    inv Hmrel3_le. tac.
  }
Qed.

Lemma foo_refinement globals funcs
      (Hwf_src: Prog.wf (Prog.mk globals (funcs ++ [foo_tgt])))
      (Hwf_tgt: Prog.wf (Prog.mk globals (funcs ++ [foo_tgt]))):
  behave_prog lts (Prog.mk globals (funcs ++ [foo_tgt])) <1=
  behave_prog lts (Prog.mk globals (funcs ++ [foo_src])).
Proof.
  apply adequacy.
  apply sim_prog.
  apply list_forall2_app.
  - unfold Prog.wf in *.
    revert Hwf_tgt. revert Hwf_src.
    generalize funcs at 1 2 7 8. intro X. induction X; constructor; tac.
    + split; tac.
      repeat intro. apply identity_sim_local_func; auto.
      * eapply FDecl.wf_mon; eauto.
      * eapply FDecl.wf_mon; eauto.
    + exploit IHX; eauto.
  - repeat constructor. by apply foo_sim_local_func_aux.
Qed.
