Require Import Equality.
Require Import Relation_Operators.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import Events.
Require Import Behaviors.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import MemoryRelation.
Require Import StateRelation.
Require Import SimulationLocal.
Require Import SimulationProgWeak.
Require Import Adequacy.
Require Import IdentityCompiler.

Import SyntaxNotations.

Program Definition foo_src :=
  FUNCTION %f"foo" WITH [%p"h"]
  LOCALS [%p"p"; %i"a"] BEGIN [
    %p"p" ::= MALLOC 1;
    123 --> %p"p";
    CALL %f"bar" WITH [];
    %i"a" <-- %p"p";
    CALL %f"hash_put" WITH [%p"h"; %p"p"; %i"a"]
  ] END
.
Lemma foo_src_wf globals: FDecl.wf globals foo_src.
Proof. wftac. Qed.

Program Definition foo_tgt :=
  FUNCTION %f"foo" WITH [%p"h"]
  LOCALS [%p"p"; %i"a"] BEGIN [
    %p"p" ::= MALLOC 1;
    123 --> %p"p";
    CALL %f"bar" WITH [];
    %i"a" <-- %p"p";
    CALL %f"hash_put" WITH [%p"h"; %p"p"; 123]
  ] END
.
Lemma foo_tgt_wf globals: FDecl.wf globals foo_tgt.
Proof. wftac. Qed.

Ltac vartac :=
  rewrite ? Env.gsspec in *;
  repeat
    (match goal with
       | [H: (_, _) = (_, _) |- _] =>
         inv H
       | [H: %f ?a = %f ?a |- _] =>
         clear H
       | [H: %f ?a = %f ?b |- _] =>
         by apply ident_of_string_inject in H
       | [|- context[Var.eq ?a ?a]] =>
         destruct (Var.eq a a); try done
       | [|- context[Var.eq ?a ?b]] =>
         destruct (Var.eq a b); try done
       | [H: context[Var.eq ?a ?a] |- _] =>
         destruct (Var.eq a a); try done
       | [H: context[Var.eq ?a ?b] |- _] =>
         destruct (Var.eq a b); try done
     end).

Lemma foo_sim_local_func_aux sst_src sst_tgt (mrel_init:MemRel.t)
      (Hmrel_init_globals: mrel_genv mrel_init sst_src.(genv) sst_tgt.(genv)):
  sim_local_func_aux lts lts sst_src sst_tgt mrel_init foo_src foo_tgt.
Proof.
  unfold foo_src, foo_tgt. repeat intro.
  exploit mrel_bind_params; eauto. intros. simtac.
  destruct argvals_src as [|argval1_src argvals_src]; simtac.
  destruct argvals_tgt as [|argval1_tgt argvals_tgt]; simtac.
  inv Hargvals. inv H6. vgtac.des. inv H. rename H0 into Hlenv1.
  exists WF.elt.

  (* 0. initialize *)
  pfold. constructor; simtac. intros.
  destruct cfg3_tgt as [st3_tgt|]; simtac.
  Focus 2.
  { unfold sem_instr in H. simtac. inv H; simtac. }
  Unfocus.
  unfold sem_instr in Htgt. simtac.
  exploit mrel_Env_init; eauto. intros. simtac. vgtac.des.
  eexists. split; [apply rt_refl|].
  exists WF.elt. eexists. split.
  { apply indexed_step_step. simtac.
    instantiate (1 := Some _). simtac.
    unfold sem_instr. simtac.
    repeat (eapply bindPP_right; eauto; tac).
  }
  inv H. inv Henv. inv Henv0. inv Hr1. inv Henv. inv Henv0.
  exploit (H0 (Typ.int, (%f"a"))); eauto; vartac. intro. simtac. inv H. inv H1.
  eexists. vgtac.splits; eauto; simtac.
  { reflexivity. }
  { constructor. simtac. }
  left.

  (* 1. malloc *)
  exploit MemRel.alloc_src; eauto; [by unfold Mem.alloc|]. instantiate (1 := 1). intros [Hmrel13 Hmrel2]. simtac.
  exploit MemRel.alloc_tgt; eauto; [by unfold Mem.alloc|]. instantiate (1 := 1). intros [Hmrel23 Hmrel3]. simtac.
  rewrite Hmrel23 in Hmrel13. clear Hmrel2 Hmrel23.

  pfold. apply sim_local_step; repeat intro; simtac.

  destruct cfg3_tgt as [st1_tgt|]; simtac.
  Focus 2.
  { unfold sem_instr in H. simtac.
    inv H; simtac. inv Hr1; simtac.
  }
  Unfocus.
  unfold sem_instr in Htgt. simtac.
  eexists. split; [apply rt_refl|].
  exists WF.elt. eexists. split.
  { apply indexed_step_step. simtac.
    instantiate (1 := Some _). simtac.
    unfold sem_instr. simtac.
    repeat (eapply bindPP_right; eauto).
    unfold Mem.val2int. simtac. right. simtac.
  }
  eexists. split; eauto. split; simtac. left.

  (* 2. store *)
  destruct
    (Mem.store (Value.ptr (Mem.nextb mem_src) Int.zero) 123
               (Mem.mk (Pos.succ (Mem.nextb mem_src))
                       (PTree.set (Mem.nextb mem_src) (Mem.Block.init 1) mem_src)))
  eqn:Hstore.
  Focus 2.
  pfold. eapply sim_local_error.
  { apply rt_step. simtac.
    instantiate (1 := None). simtac.
    unfold sem_instr. split; simtac.
    repeat (eapply bindPP_right; eauto; simtac).
    by rewrite Hstore.
  }
  { simtac. }
  Unfocus.

  exploit MemRel.store_src; eauto; simtac.
  intro. simtac.
  exploit MemRel.store_tgt; eauto.
  { instantiate (3 := (Value.ptr (Mem.nextb mem_tgt) Int.zero)).
    unfold Mem.val2ptr, Mem.normalize. simtac.
  }
  { simtac. }
  { instantiate (2 := 123). unfold Mem.store_aux. simtac. }
  intro. simtac.

  unfold Mem.store_aux in H. simtac.
  unfold Mem.Block.store in Hcond. simtac.

  pfold. apply sim_local_step; repeat intro; simtac.

  destruct cfg3_tgt as [st1_tgt|]; simtac.
  Focus 2.
  { unfold sem_instr in H. simtac.
    inv H; tac; vartac. inv Hr2; simtac.
    rewrite H3 in Hr0. simtac.
  }
  Unfocus.
  unfold sem_instr in Htgt. simtac. vartac. simtac.

  eexists. split; [apply rt_refl|].
  exists WF.elt. eexists. split.
  { apply indexed_step_step. simtac. unfold sem_instr. simtac.
    instantiate (1 := Some _). simtac. vartac.
    repeat (try (eapply bindPP_right; eauto); tac).
    rewrite Hstore. simtac.
  }
  eexists. split.
  { rewrite H1. eauto. }
  split; [|left].
  { constructor; simpl; eauto. }
  unfold Mem.Block.store in Hcond1. simtac.

  (* 3. call *)
  pfold. eapply sim_local_call; repeat intro; simtac.
  { apply rt_refl. }
  { simtac. }
  { eauto. }
  { reflexivity. }
  { simtac. }
  { simtac. }
  { done. }
  { simtac. constructor. }
  simtac. exists WF.elt. left.
  destruct st3_src as [ctx3_src stack3_src mem3_src].
  destruct st3_tgt as [ctx3_tgt stack3_tgt mem3_tgt]. simtac.

  (* 4. load *)
  pfold. eapply sim_local_step; repeat intro; simtac.

  assert (X: (MemRel.tgt_private mrel3) ! (Mem.nextb mem_tgt) = Some (Mem.Block.mk true true 1 None (ZMap.set 0 (Value.int (Int.repr 123)) (ZMap.init (Value.int Int.zero))))).
  { inv Hmrel3_le. simtac. apply Htgt_private. simtac. }
  assert (Y: mem3_tgt ! (mem_tgt.(Mem.nextb)) = (MemRel.tgt_private mrel3) ! (Mem.nextb mem_tgt)).
  { erewrite get_tgt; eauto. }
  exploit MemRel.load_tgt; eauto.
  { instantiate (2 := (Value.ptr mem_tgt.(Mem.nextb) Int.zero)).
    unfold Mem.val2ptr. simtac. rewrite Y, X. simtac.
  }
  intro Z.

  assert (W: (MemRel.src_private mrel3) ! (Mem.nextb mem_src) = Some (Mem.Block.mk true true 1 None (ZMap.set 0 (Value.int (Int.repr 123)) (ZMap.init (Value.int Int.zero))))).
  { inv Hmrel3_le. simtac. apply Hsrc_private. simtac. }
  assert (A: mem3_src ! (mem_src.(Mem.nextb)) = (MemRel.src_private mrel3) ! (Mem.nextb mem_src)).
  { erewrite get_src; eauto. }
  exploit MemRel.load_src; eauto.
  { instantiate (2 := (Value.ptr mem_src.(Mem.nextb) Int.zero)).
    unfold Mem.val2ptr. simtac. rewrite A, W. simtac.
  }
  intro B.

  destruct cfg3_tgt as [st1_tgt|]; simtac.
  Focus 2.
  { unfold sem_instr in H. simtac.
    inv H; tac; vartac. rewrite Z in Hr2.
    unfold Mem.load_aux in Hr2. rewrite X in Hr2. inv Hr2; simtac.
  }
  Unfocus.
  unfold sem_instr in Htgt. simtac. destruct Hr2; simtac.
  rewrite Z in Hcond1. inv Hcond1.
  unfold Mem.load_aux in H8. simtac. unfold Mem.Block.load in H9. simtac.
  assert (C: (Int.unsigned Int.zero) = 0); [|rewrite C in *].
  { done. }
  assert (D: (Int.unsigned 1) = 1).
  { done. }
  rewrite ZMap.gsspec in *. simtac.

  eexists. split; [apply rt_refl|]. exists WF.elt. eexists. split.
  { apply indexed_step_step. simtac.
    instantiate (1 := Some _). simtac.
    unfold sem_instr. simtac.
    repeat (try (eapply bindPP_right; eauto); simtac).
    rewrite B. unfold Mem.load_aux. rewrite W.
    unfold Mem.Block.load. simtac. rewrite Hcond3. rewrite ZMap.gsspec, C. simtac.
  }
  eexists. split; [reflexivity|]. split; simtac.
  left.

  (* 5. call *)
  exploit MemRel.public; eauto.
  { constructor; simtac. intros. rewrite ZMap.gsspec.
    destruct (ZIndexed.eq (Int.unsigned i) 0); simtac; try constructor.
    rewrite ZMap.gi. constructor.
  }
  intros [Hmrel45 Hmrel5].

  pfold. eapply sim_local_call; repeat intro; simtac.
  { apply rt_refl. }
  { simtac. }
  { eauto. }
  { auto. }
  { simtac. }
  { simtac. }
  { rewrite peq_false; [|by intro X; apply ident_of_string_inject in X].
    simtac. rewrite Hcond0. simtac.
  }
  { simtac. rewrite peq_false; [|by intro X; apply ident_of_string_inject in X].
    simtac. rewrite Hcond. simtac.
  }
  { repeat constructor; simtac.
    { eapply MemRel.incr_rel_value_aux; simtac.
      { instantiate (1 := mrel_entry).
        intros.
        inv Hmrel3_le. inv Hequiv. simtac. apply Hmbinrel in PR.
        unfold MemEquiv.add, MemEquiv.to_relation. simtac.
        destruct (peq x0 mem_src.(Mem.nextb)); simtac.
        inv Hst3_mem. simtac. exploit Hsrc_private0; eauto.
        { instantiate (1 := mem_src.(Mem.nextb)). by rewrite PR. }
        intro X. by rewrite W in X.
      }
      { instantiate (1 := mem_src).
        inv Hmrel45. simtac. rewrite <- Hsrc.
        inv Hmrel3_le. inv Hequiv. simtac. rewrite <- Hsrc0.
        inv H1. simtac. rewrite <- Hsrc1.
        inv Hmrel13. simtac. rewrite <- Hsrc2.
        inv Hmrel_entry. reflexivity.
      }
      { instantiate (1 := mem_tgt).
        inv Hmrel45. simtac. rewrite <- Htgt.
        inv Hmrel3_le. inv Hequiv. simtac. rewrite <- Htgt0.
        inv H5. simtac. rewrite <- Htgt1.
        inv Hmrel13. simtac. rewrite <- Htgt2.
        inv Hmrel_entry. reflexivity.
      }
      { inv Hst3_mem. destruct Hpublic. simtac.
        repeat intro. exploit Haddress; eauto. intro. simtac.
        exists x. split.
        { unfold MemEquiv.add, MemEquiv.to_relation. simtac.
          destruct (peq b_src mem_src.(Mem.nextb)); simtac.
        }
        { rewrite H7. eexists. split; eauto. }
      }
      { auto. }
    }
    { unfold MemEquiv.add, MemEquiv.to_relation. simtac. by rewrite peq_true.
    }
  }
  simtac. exists WF.elt. left.
  destruct st3_src as [ctx5_src stack5_src mem5_src].
  destruct st3_tgt as [ctx5_tgt stack5_tgt mem5_tgt]. simtac.

  (* 6. return *)
  pfold. eapply sim_local_return; eauto; simtac.
  { reflexivity. }
  rewrite <- Hmrel3_le0.
  constructor; simtac.
  { rewrite <- Hmrel45.
    inv Hmrel3_le. rewrite <- Hequiv.
    rewrite <- H5, <- H1. auto.
  }
  { repeat intro. simtac. unfold PTree.elt_eq.
    destruct (peq i (Mem.nextb mem_src)); simtac.
    { inv Hmrel_entry. inv Hmrel3_le. inv Hsrc_wf. exploit _nextb; eauto.
      intro X. apply Plt_ne in X. simtac.
    }
    inv Hmrel3_le. simtac. erewrite Hsrc_private; eauto. simtac.
  }
  { repeat intro. simtac. unfold PTree.elt_eq.
    destruct (peq i (Mem.nextb mem_tgt)); simtac.
    { inv Hmrel_entry. inv Hmrel3_le. inv Htgt_wf. exploit _nextb; eauto.
      intro X. apply Plt_ne in X. simtac.
    }
    inv Hmrel3_le. simtac. erewrite Htgt_private; eauto. simtac.
  }
Qed.

Lemma foo_refinement globals funcs
      (Hwf_src: Prog.wf (Prog.mk globals (funcs ++ [foo_tgt])))
      (Hwf_tgt: Prog.wf (Prog.mk globals (funcs ++ [foo_tgt]))):
  behave_prog lts (Prog.mk globals (funcs ++ [foo_tgt])) <1=
  behave_prog lts (Prog.mk globals (funcs ++ [foo_src])).
Proof.
  apply adequacy.
  apply sim_prog.
  apply list_forall2_app.
  - unfold Prog.wf in *.
    revert Hwf_tgt. revert Hwf_src.
    generalize funcs at 1 2 7 8. intro X. induction X; constructor; tac.
    + split; tac.
      repeat intro. apply identity_sim_local_func; auto.
      * eapply FDecl.wf_mon; eauto.
      * eapply FDecl.wf_mon; eauto.
    + exploit IHX; eauto.
  - repeat constructor. by apply foo_sim_local_func_aux.
Qed.
