Require Import Equality.
Require Import Relation_Operators.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import MemoryC.
Require Import Events.
Require Import Behaviors.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import SemanticsC.
Require Import MemoryRelationC.
Require Import StateRelationC.
Require Import SimulationLocalC.

Lemma init_globals_mrel_genv_aux
      genv1 mem1 tmem1
      (mrel_init1:MemRelC.t)
      (Hgenv1: mrel_genv mrel_init1 genv1 genv1)
      (Hmrel1: mrel_init1 mem1 tmem1)
      (Hnextb: mem1.(Mem.nextb) = tmem1.(MemC.nextb))
      globals genv2 tmem2 (Hinit: init_globals globals genv1 tmem1 (genv2, tmem2)):
  exists (mrel_init2:MemRelC.t) (mem2:Mem.t),
    mrel_genv mrel_init2 genv2 genv2 /\
    Semantics.init_globals globals genv1 mem1 (genv2, mem2) /\
    mrel_init2 mem2 tmem2 /\
    mem2.(Mem.nextb) = tmem2.(MemC.nextb).
Proof.
  generalize dependent mrel_init1.
  generalize dependent tmem2.
  generalize dependent genv2.
  generalize dependent tmem1.
  generalize dependent genv1.
  generalize dependent mem1.
  induction globals; intros.
  { inv Hinit. eexists. eexists. vgtac.splits; eauto. constructor. }
  destruct a as [gvar size]. inv Hinit.
  exploit MemRelC.alloc_public; eauto.
  { by inv Hmrel1. }
  { by inv Hmrel1. }
  intro. des.
  exploit Mem.alloc_inv; eauto. intro. des. subst.
  inv H12. apply MemC.alloc_inv in Halloc. des. subst.
  inv Hrealize. inv Hblk2.
  exploit IHglobals; try apply H23; eauto.
  { simpl. congruence. }
  { repeat intro. tac.
    destruct (peq gvar0 gvar); subst.
    - inv Hb_src. eexists. split; eauto.
      unfold MemEquivC.add, MemEquivC.to_relation. simpl.
      destruct (peq (MemC.nextb tmem1) (Mem.nextb mem1)); auto.
      congruence.
    - exploit Hgenv1; eauto. intro. des.
      unfold ident in Hb_src. rewrite Hb_src in H2. inv H2.
      eexists. split; eauto.
      unfold MemEquivC.add, MemEquivC.to_relation. simpl.
      destruct (peq b_tgt (Mem.nextb mem1)); auto.
      congruence.
  }
  intros [mrel_init2 [mem2 [Hgenv2 [Hinit2 [Hmem2 Hnextb2]]]]].
  eexists. eexists. vgtac.splits; eauto. econstructor; eauto.
  congruence.
Qed.

Lemma init_globals_mrel_genv
      globals genv tmem (Hinit: init_globals globals GEnv.empty MemC.init (genv, tmem)):
  exists (mrel_init:MemRelC.t) (mem:Mem.t),
    mrel_genv mrel_init genv genv /\
    Semantics.init_globals globals GEnv.empty Mem.init (genv, mem) /\
    mrel_init mem tmem.
Proof.
  exploit init_globals_mrel_genv_aux; eauto.
  { repeat intro. by rewrite PTree.gempty in *. }
  { instantiate (1 := Mem.init).
    instantiate (1 := MemRelC.mk (MemEquivC.add 1%positive 1%positive MemEquivC.empty) (PTree.empty _) (PTree.empty _) Mem.init MemC.init).
    constructor; simpl; tac.
    - apply MemEquivC.add_wf.
      + intros [? ?]. inv H.
      + intros [? ?]. inv H.
      + apply MemEquivC.empty_wf.
    - apply Mem.init_wf.
    - apply MemC.init_wf.
    - repeat intro. by rewrite PTree.gempty.
    - repeat intro. by rewrite PTree.gempty.
    - repeat intro. by rewrite PTree.gempty in H1.
    - repeat intro. by rewrite PTree.gempty in H1.
    - assert (H11: (MemEquivC.add 1%positive 1%positive MemEquivC.empty) 1%positive 1%positive).
      { unfold MemEquivC.to_relation, MemEquivC.add. simpl.
        by rewrite peq_true.
      }
      constructor; auto.
      + repeat intro. unfold Mem.init, MemC.init in *. rewrite <- Mem.blocks_unsimpl, <- MemC.blocks_unsimpl in *.
        rewrite PTree.gsspec in *. destruct (peq b_src 1); subst.
        * exists 1%positive. split; auto.
          rewrite PTree.gss. eexists. split; eauto.
          by inv Hblk.
        * by rewrite PTree.gempty in Hblk.
      + intros. inv Hi.
        destruct (peq si 1); inv H0.
        unfold Mem.init. rewrite <- Mem.blocks_unsimpl.
        rewrite PTree.gss. constructor; auto.
        simpl. intros. rewrite ? ZMap.gi. constructor.
  }
  { reflexivity. }
  { intro. des. eexists. eexists. vgtac.splits; eauto. }
Qed.

Require Import vgtac.

Lemma list_forall2_rev {T U:Type} (p: T -> U -> Prop) (l1:list T) (l2:list U):
  list_forall2 p (rev l1) (rev l2) -> list_forall2 p l1 l2.
Proof.
  intro X. dependent induction X.
  - apply (eq_apply (@rev _)) in x0. rewrite rev_involutive in x0. subst.
    apply (eq_apply (@rev _)) in x. rewrite rev_involutive in x. subst.
    constructor.
  - apply (eq_apply (@rev _)) in x0. rewrite rev_involutive in x0. subst.
    apply (eq_apply (@rev _)) in x. rewrite rev_involutive in x. subst.
    tac. apply list_forall2_app; tac.
    { apply IHX; by rewrite rev_involutive. }
    repeat constructor; auto.
Qed.

Lemma sim_prog
      globals funcs_src funcs_tgt
      (Hfuncs:
         list_forall2
           (fun func_src func_tgt =>
              <<Hfunc_id: func_src.(FDecl.id) = func_tgt.(FDecl.id)>> /\
              <<Hfunc_sim: forall sst_src sst_tgt
                                  (Hgenv_src: list_subset (map fst globals) (PTree.keys sst_src.(Semantics.genv)))
                                  (Hgenv_tgt: list_subset (map fst globals) (PTree.keys sst_tgt.(genv)))
                                  (Hfenv_src: sst_src.(Semantics.fenv) = fold_left
                                                                 (fun (acc : PTree.t FDecl.t) (fdecl : FDecl.t) =>
                                                                    PTree.set (FDecl.id fdecl) fdecl acc) funcs_src (PTree.empty _))
                                  (Hfenv_tgt: sst_tgt.(fenv) = fold_left
                                                                 (fun (acc : PTree.t FDecl.t) (fdecl : FDecl.t) =>
                                                                    PTree.set (FDecl.id fdecl) fdecl acc) funcs_tgt (PTree.empty _))
                                  (mrel_init:MemRelC.t)
                                  (Hmrel_init: mrel_genv mrel_init sst_src.(Semantics.genv) sst_tgt.(genv)),
                             sim_local_func lts ltsC sst_src sst_tgt mrel_init func_src func_tgt>>)
           funcs_src funcs_tgt):
  sim_prog lts ltsC (Prog.mk globals funcs_src) (Prog.mk globals funcs_tgt).
Proof.
  repeat intro. tac. unfold load_prog in *. tac.
  inv Hprog_tgt. clear Hprog_tgt. destruct Hr2. destruct v1 as [genv1 mem1]. simpl in *. subst.
  exploit init_globals_mrel_genv; eauto. intros [mrel_init [smem1 [Hgenv1 [Hmrel_init_globals Hmrel_init_mem]]]].
  eexists. eexists. exists mrel_init.
  splits; [eapply bindPP_right; eauto| | |]; eauto; tac.
  repeat intro. tac. rewrite <- FDecl_find in *.
  revert Hfunc_src. revert func_src.
  rewrite <- (rev_involutive funcs_src) in Hfuncs at 1.
  rewrite <- (rev_involutive funcs_tgt) in Hfuncs at 1.
  apply list_forall2_rev in Hfuncs.

  generalize dependent Hfuncs.
  generalize (rev funcs_tgt).
  generalize (rev funcs_src).
  intros l1 l2 X. dependent induction X; intros; tac.
  unfold FDecl.byid in *. rewrite <- H. destruct (peq fid a1.(FDecl.id)); tac. 
  { eexists. split; tac. apply H0; tac.
    - repeat intro. eapply init_globals_keys; eauto.
    - repeat intro. eapply init_globals_keys; eauto.
  }
  { exploit IHX; eauto. }
Qed.
