Require Import RelationClasses.
Require Import Axioms.
Require Import Coqlib.
Require Import Kanglib.
Require Import vgtac.
Require Import paco.
Require Import Integers.
Require Import Maps.
Require Import Monad.
Require Import AST.
Require Import Values.
Require Import Memory.
Require Import MemoryRelation.
Require Import Syntax.
Require Import Semantics.

Ltac tac :=
  repeat
    (simpl in *; unfold SomeP, NoneP, singletonP in *; subst; pclearbot;
     (try done);
     (try basictac_aux);
     (try destac_aux);
     (try monadtac_aux);
     (try maptac_aux);
     (try memtac_aux);
     (try match goal with
            | [st:State |- _] =>
              let lenv := fresh "lenv" in
              let istrs := fresh "istrs" in
              let stack := fresh "stack" in
              let mem := fresh "mem" in
              destruct st as [[lenv istrs] stack mem]
          end)).

Definition mrel_genv (mrel:MemEquiv.t) (genv_src genv_tgt:GEnv.t): Prop :=
  forall gvar b_src (Hb_src: PTree.get gvar genv_src = Some b_src),
  exists b_tgt,
    <<Hb_tgt: PTree.get gvar genv_tgt = Some b_tgt>> /\
    <<Hb: mrel b_src b_tgt>>.

Definition mrel_env (mrel:MemEquiv.t) mem_src mem_tgt env_src env_tgt: Prop :=
  forall var val_src (Hval_src: Env.get var env_src = Some val_src),
  exists val_tgt,
    <<Hval_tgt: Env.get var env_tgt = Some val_tgt>> /\
    <<Hval: MemEquiv.rel_value mrel mem_src mem_tgt val_src val_tgt>>.

Lemma mrel_genv_le
      (mrel1 mrel2: MemEquiv.t) genv_src genv_tgt
      (Hmrel1: mrel_genv mrel1 genv_src genv_tgt)
      (Hmrel12: mrel1 <2= mrel2):
  mrel_genv mrel2 genv_src genv_tgt.
Proof.
  repeat intro. exploit Hmrel1; eauto. intro. tac.
  eexists. split; eauto.
Qed.

Lemma mrel_env_le
      lenv_src lenv_tgt mem1_src mem1_tgt mem2_src mem2_tgt (mrel1 mrel2: MemRel.t)
      (Hle: MemRel.le mrel1 mrel2)
      (Hmrel1: mrel1 mem1_src mem1_tgt)
      (Hmrel2: mrel2 mem2_src mem2_tgt)
      (Hlenv: mrel_env mrel1 mem1_src mem1_tgt lenv_src lenv_tgt):
  mrel_env mrel2 mem2_src mem2_tgt lenv_src lenv_tgt.
Proof.
  repeat intro. exploit Hlenv; eauto. intro. tac.
  eexists. split; eauto.
  eapply MemRel.incr_rel_value; eauto.
Qed.

Lemma mrel_Env_set
      var v_src v_tgt lenv_src lenv_tgt mem_src mem_tgt (mrel:MemRel.t)
      (Hv: MemEquiv.rel_value mrel mem_src mem_tgt v_src v_tgt)
      (Hlenv: mrel_env mrel mem_src mem_tgt lenv_src lenv_tgt):
  mrel_env mrel mem_src mem_tgt (Env.set var v_src lenv_src) (Env.set var v_tgt lenv_tgt).
Proof.
  repeat intro. rewrite Env.gsspec in *.
  destruct (Var.eq var0 var); tac.
  - eexists. split; eauto.
  - by apply Hlenv.
Qed.

Lemma mrel_Env_empty
      mrel mem_src mem_tgt:
      mrel_env mrel mem_src mem_tgt Env.empty Env.empty.
Proof.
  repeat intro.
  rewrite Env.gempty in Hval_src; tac.
Qed.

Lemma mrel_has_type
      v_src v_tgt typ
      (mrel:MemRel.t) (mem_src mem_tgt:Mem.t)
      (Hmrel: mrel mem_src mem_tgt)
      (Hv: MemEquiv.rel_value mrel mem_src mem_tgt v_src v_tgt)
      (Hv_src: Value.has_type v_src typ):
  Value.has_type v_tgt typ.
Proof.
  inv Hv.
Qed.

Lemma mrel_exp_has_type
      exp typ
      genv_src genv_tgt env_src env_tgt
      (mrel:MemRel.t) (mem_src mem_tgt:Mem.t)
      (Hmrel_genv: mrel_genv mrel genv_src genv_tgt)
      (Hmrel_env: mrel_env mrel mem_src mem_tgt env_src env_tgt)
      (Htyp: Exp.has_type (PTree.keys genv_src) (Env.keys env_src) exp = Some typ):
  Exp.has_type (PTree.keys genv_tgt) (Env.keys env_tgt) exp = Some typ.
Proof.
  eapply Exp.has_type_mon; eauto.
  - repeat intro. apply PTree.gkspec in H.
    destruct (genv_src ! a) eqn:Ha; tac.
    exploit Hmrel_genv; eauto. intro. des.
    apply PTree.gkspec. by rewrite Hb_tgt.
  - repeat intro. apply Env.gkspec in H.
    destruct (Env.get a env_src) eqn:Ha; tac.
    exploit Hmrel_env; eauto. intro. des.
    apply Env.gkspec. by rewrite Hval_tgt.
Qed.

Lemma bind_params_env
      params vals mem env1 env2 env1'
      (Hbind : bind_params params vals mem env1 = Some env1'):
  exists env2', bind_params params vals mem env2 = Some env2'.
Proof.
  revert vals env1 env2 env1' Hbind; induction params; tac.
  - destruct vals; tac; eauto.
  - destruct vals; tac; intros.
    destruct (Value.has_type t0 t); tac; eauto.
Qed.

Lemma mrel_bind_params_types
      (params_src params_tgt : list Var.t) argvals_src argvals_tgt
      (mrel:MemRel.t) (mem_src mem_tgt:Mem.t)
      env1_src env1_tgt env2_src
      (Hmrel: mrel mem_src mem_tgt)
      (ENV: mrel_env mrel mem_src mem_tgt env1_src env1_tgt)
      (VEQ: list_forall2 (MemEquiv.rel_value mrel mem_src mem_tgt) argvals_src argvals_tgt)
      (TYPES: list_forall2 (fun v_src v_tgt => fst v_src = fst v_tgt) params_src params_tgt)
      (BINDP: bind_params params_src argvals_src mem_src env1_src = Some env2_src):
  exists env2_tgt,
    bind_params params_tgt argvals_tgt mem_tgt env1_tgt = Some env2_tgt.
Proof.
  generalize dependent env1_tgt.
  generalize dependent env1_src.
  generalize dependent env2_src.
  generalize dependent argvals_tgt.
  generalize dependent argvals_src.
  generalize dependent params_tgt.
  induction params_src; intros; tac.
  - inv VEQ; inv TYPES; tac; eauto.
  - inv VEQ. clear VEQ. inv TYPES; clear TYPES; tac.
    exploit IHparams_src; eauto.
    + apply mrel_Env_set; eauto.
    + intro. tac.
      exploit mrel_has_type; eauto; intro X; rewrite X.
      eapply bind_params_env; eauto.
Qed.

Lemma mrel_bind_params_aux
      (params:list Var.t) argvals_src argvals_tgt
      (mrel:MemRel.t) (mem_src mem_tgt:Mem.t)
      env1_src env1_tgt env2_src
      (Hmrel: mrel mem_src mem_tgt)
      (ENV: mrel_env mrel mem_src mem_tgt env1_src env1_tgt)
      (VEQ: list_forall2 (MemEquiv.rel_value mrel mem_src mem_tgt) argvals_src argvals_tgt)
      (BINDP: bind_params params argvals_src mem_src env1_src = Some env2_src):
  exists env2_tgt,
    <<Henv_tgt: bind_params params argvals_tgt mem_tgt env1_tgt = Some env2_tgt>> /\
    <<Henv: mrel_env mrel mem_src mem_tgt env2_src env2_tgt>>.
Proof.
  generalize dependent env1_tgt.
  generalize dependent env1_src.
  generalize dependent env2_src.
  generalize dependent argvals_tgt.
  generalize dependent argvals_src.
  induction params; intros; tac.
  - inv VEQ. eexists. split; eauto.
  - inv VEQ. clear VEQ. tac.
    exploit IHparams; eauto.
    + apply mrel_Env_set; eauto.
    + intro. tac. eexists. split; eauto.
      rewrite H1. exploit mrel_has_type; eauto.
      intro X. by rewrite X.
Qed.

Lemma mrel_bind_params
      (params:list Var.t) argvals_src argvals_tgt
      (mrel:MemRel.t) (mem_src mem_tgt:Mem.t)
      env_src
      (Hmrel: mrel mem_src mem_tgt)
      (VEQ: list_forall2 (MemEquiv.rel_value mrel mem_src mem_tgt) argvals_src argvals_tgt)
      (BINDP: bind_params params argvals_src mem_src Env.empty = Some env_src):
  exists env_tgt,
    <<Henv_tgt: bind_params params argvals_tgt mem_tgt Env.empty = Some env_tgt>> /\
    <<Henv: mrel_env mrel mem_src mem_tgt env_src env_tgt>>.
Proof.
  eapply mrel_bind_params_aux; eauto.
  repeat intro. by rewrite Env.gempty in Hval_src.
Qed.  

Lemma mrel_Env_init
      lenv1_src lenv1_tgt lenv2_tgt mem_src mem_tgt (mrel:MemRel.t) params
      (Hmrel: mrel mem_src mem_tgt)
      (Hmrel1: mrel_env mrel mem_src mem_tgt lenv1_src lenv1_tgt)
      (Hlenv2_tgt: Env.init params lenv1_tgt lenv2_tgt):
  exists lenv2_src,
    <<Hlenv2_src: Env.init params lenv1_src lenv2_src>> /\
    <<Hlenv2: mrel_env mrel mem_src mem_tgt lenv2_src lenv2_tgt>>.
Proof.
  generalize dependent lenv2_tgt.
  generalize dependent lenv1_tgt.
  generalize dependent lenv1_src.
  induction params; intros; tac.
  - inv Hlenv2_tgt. exists lenv1_src. split; tac. constructor.
  - inv Hlenv2_tgt.
    { exploit IHparams; eauto. intros. tac.
      eexists. split; [econstructor; eauto|].
      repeat intro. rewrite Env.gsspec in *.
      destruct (Var.eq var0 (Typ.int, var)); tac.
      * eexists. split; eauto.
        constructor.
      * by apply H0.
    }
    { exploit IHparams; eauto. intros. tac.
      eexists. split; [econstructor; eauto|].
      repeat intro. rewrite Env.gsspec in *.
      destruct (Var.eq var0 (Typ.ptr, var)); tac.
      * eexists. split; eauto.
        constructor. inv Hmrel. inv Hpublic.
      * by apply H0.
    }
Qed.

Lemma mrel_bexp
      (e:Exp.bt) v_src
      env_src env_tgt genv_src genv_tgt (mem_src mem_tgt:Mem.t) (mrel:MemRel.t)
      (Hmrel_genv: mrel_genv mrel genv_src genv_tgt)
      (Hmrel_env: mrel_env mrel mem_src mem_tgt env_src env_tgt)
      (Hmrel: mrel mem_src mem_tgt)
      (Henv_src_wf: Env.wf (Env.keys env_src) env_src)
      (Henv_tgt_wf: Env.wf (Env.keys env_tgt) env_tgt)
      (He_wf: Exp.has_type (PTree.keys genv_src) (Env.keys env_src) e)
      (Hv_src: sem_bexp genv_src env_src e = Some v_src):
  exists v_tgt,
    <<Hv_tgt: sem_bexp genv_tgt env_tgt e = Some v_tgt>> /\
    <<Hv: MemEquiv.rel_value mrel mem_src mem_tgt v_src v_tgt>>.
Proof.
  destruct e; tac.
  - eexists. split; eauto. constructor.
  - exploit Hmrel_genv; eauto. intro. tac. rewrite H.
    eexists. split; tac.
    inv Hmrel. destruct Hpublic. tac.
    exploit Hpublic; try apply H0; eauto. intro X. inv X. clear X.
    destruct s.(Mem.Block.address) eqn:Haddr_src; tac.
    + exploit Haddress; eauto. intro. tac.
      exploit (MemEquiv.src_unique p x x0); eauto. intro. tac.
      rewrite H4 in H3. inv H3. clear H3.
      constructor. auto.
    + econstructor; eauto.
  - exploit Hmrel_env; eauto. intro. tac. rewrite H.
    eapply for_all_spec in Henv_tgt_wf; eauto. rewrite H in Henv_tgt_wf. tac.
    rewrite Henv_tgt_wf. eexists. split; tac.
    apply Env.gkspec. by rewrite H.
Qed.

Lemma mrel_bexp_list
      (es:list Exp.bt) vs_src
      env_src env_tgt genv_src genv_tgt (mem_src mem_tgt:Mem.t) (mrel:MemRel.t)
      (Hmrel_genv: mrel_genv mrel genv_src genv_tgt)
      (Hmrel_env: mrel_env mrel mem_src mem_tgt env_src env_tgt)
      (Hmrel: mrel mem_src mem_tgt)
      (Henv_src_wf: Env.wf (Env.keys env_src) env_src)
      (Henv_tgt_wf: Env.wf (Env.keys env_tgt) env_tgt)
      (He_wf: for_all (fun (e:Exp.bt) => Exp.has_type (PTree.keys genv_src) (Env.keys env_src) e) es)
      (Hv_src: sem_bexp_list genv_src env_src es = Some vs_src):
  exists vs_tgt,
    <<Hvs_tgt: sem_bexp_list genv_tgt env_tgt es = Some vs_tgt>> /\
    <<Hvs: list_forall2 (MemEquiv.rel_value mrel mem_src mem_tgt) vs_src vs_tgt>>.
Proof.
  generalize dependent vs_src.
  generalize dependent es. induction es; intros; tac.
  - exists nil. split; auto. constructor.
  - exploit mrel_bexp; eauto. intro. tac.
    exploit IHes; eauto. intro. tac.
    rewrite H1. rewrite H3. tac.
    eexists. split; eauto. constructor; auto.
Qed.

Definition typ_sig_eq_dec (t1 t2 t1' t2' : Typ.t) : 
  {(t1, t2) = (t1', t2')} + {(t1, t2) <> (t1', t2')}.
Proof. repeat (decide equality). Defined.

Lemma mrel_bop:
  forall (mrel: MemRel.t) (m m': Mem.t) bop t1 v1 t2 v2 v3 v1' v2' t3
         (Hbop: sem_bop m bop t1 v1 t2 v2 = Some v3)
         (Hmrel : mrel m m')
         (Hrel1 : MemEquiv.rel_value (MemRel.equiv mrel) m m' v1 v1')
         (Hrel2 : MemEquiv.rel_value (MemRel.equiv mrel) m m' v2 v2')
         (Htys : Value.has_type v1 t1 /\ Value.has_type v1' t1 /\ Value.has_type v2 t2 /\ Value.has_type v2' t2 /\
                 Bop.has_type bop t1 t2 = Some t3),
  exists v3', sem_bop m' bop t1 v1' t2 v2' = Some v3' /\ 
  MemEquiv.rel_value (MemRel.equiv mrel) m m' v3 v3'.
Proof.
  intros. tac.
  destruct bop, t1, t2, v1', v2', v3; simpl in *; tac; inv Hrel1; inv Hrel2;
  try (eexists; split; eauto; try constructor; auto; fail);
  try (eexists; split; [|econstructor]; constructor; fail);
  tac.
  - exploit (MemEquiv.src_unique si0 b b0); eauto. intro. subst b0.
    destruct (ident_eq b b); [|done]. eexists. split; eauto. constructor.
  - destruct (ident_eq b b0); subst.
    + exploit (MemEquiv.tgt_unique si si0 b0); eauto; [inv Hmrel|]. intro. subst si0.
      destruct (ident_eq si si); [|done]. inv Hbop.
      eexists. split; eauto. constructor.
    + destruct (ident_eq si si0); subst.
      * exploit (MemEquiv.src_unique si0 b b0); eauto.
        intro. subst b0. congruence.
      * inv Hmrel. destruct Hpublic.
        exploit Hpublic; try apply Hi; eauto.
        intro X. inv X. tac.
        rewrite <- Hvalid, H0, <- Hsize, H1. simpl.
        exploit Hpublic; try apply Hi0; eauto.
        intro Y. inv Y. tac. rewrite Hcond1 in H6. inv H6.
        rewrite <- Hvalid0, <- Hsize0, H3. simpl.
        destruct (ident_eq b b0); [done|].
        destruct (ident_eq i i1); [done|].
        eexists. split; eauto. constructor.
  - destruct (ident_eq si si0); subst.
    + exploit (MemEquiv.src_unique si0 b b0); eauto. intro. subst b0.
      destruct (ident_eq b b); [|done]. inv Hbop.
    + destruct (ident_eq b b0); subst.
      * exploit (MemEquiv.tgt_unique si si0 b0); eauto; [inv Hmrel|].
        intro. subst si0. congruence.
      * inv Hmrel. destruct Hpublic.
        exploit Hpublic; try apply Hi; eauto.
        intro X. inv X. tac.
        destruct (ident_eq i i1); [done|]. inv H4.
Qed.

Inductive lift2 {A B:Type} (P:forall (a:A) (b:B), Prop): forall (oa:option A) (ob:option B), Prop :=
| lift2_None ob:
    lift2 P None ob
| lift2_Some a b (HP: P a b):
    lift2 P (Some a) (Some b)
.

(* An expression can error when:
   - it accesses an undefined variable; or
   - it has a type error.
 *)
Lemma mrel_exp
      (e:Exp.t) ov_tgt
      env_src env_tgt genv_src genv_tgt (mem_src mem_tgt:Mem.t) (mrel:MemRel.t)
      (Hmrel_genv: mrel_genv mrel genv_src genv_tgt)
      (Hmrel_env: mrel_env mrel mem_src mem_tgt env_src env_tgt)
      (Hmrel: mrel mem_src mem_tgt)
      (Henv_src_wf: Env.wf (Env.keys env_src) env_src)
      (Henv_tgt_wf: Env.wf (Env.keys env_tgt) env_tgt)
      (He_wf: Exp.has_type (PTree.keys genv_src) (Env.keys env_src) e)
      (Hv_tgt: sem_exp genv_tgt env_tgt mem_tgt e ov_tgt):
  exists ov_src,
    <<Hov_src: sem_exp genv_src env_src mem_src e ov_src>> /\
    <<Hov: lift2 (MemEquiv.rel_value mrel mem_src mem_tgt) ov_src ov_tgt>>.
Proof.
  generalize dependent ov_tgt.
  generalize dependent e.
  induction e; intros; tac.
  - eexists. split; eauto.
    destruct (sem_bexp genv_src env_src bexp) as [ov_src|] eqn: Hov_src; [|by constructor].
    exploit mrel_bexp; eauto. intro. tac.
    rewrite H. constructor. auto.
  - destruct (Exp.has_type (PTree.keys genv_src) (Env.keys env_src) e1) eqn: Hty1; tac.
    destruct (Exp.has_type (PTree.keys genv_src) (Env.keys env_src) e2) eqn: Hty2; tac.
    destruct (Typ.eq t1 t); tac. destruct (Typ.eq t2 t0); tac.
    inv Hv_tgt; clear Hv_tgt.
    { exploit IHe1; eauto. intros [ov1_src [Hov1_src Hov1]].
      destruct ov1_src; inv Hov1; clear Hov1.
      eexists. split.
      - econstructor; eauto.
      - constructor.
    }
    destruct Hr2; tac.
    { exploit IHe1; eauto. intros [ov1_src [Hov1_src Hov1]].
      destruct ov1_src; inv Hov1; clear Hov1.
      - eexists. split.
        + eapply bindPP_right; eauto. left. split; eauto.
          intro X. apply H. inv HP.
        + constructor.
      - eexists. split; [|constructor].
        econstructor; eauto.
    }
    inv H0; clear H0.
    { exploit IHe1; eauto. intros [ov1_src [Hov1_src Hov1]].
      destruct ov1_src; inv Hov1; clear Hov1.
      - exploit IHe2; eauto. intros [ov2_src [Hov2_src Hov2]].
        destruct ov2_src; inv Hov2; clear Hov2.
        eexists. split.
        + eapply bindPP_right; eauto. right. split; [eapply sem_exp_type; eauto|].
          econstructor; eauto.
        + constructor.
      - eexists. split; [|constructor].
        econstructor; eauto.
    }
    destruct Hr2; tac.
    { exploit IHe1; eauto. intros [ov1_src [Hov1_src Hov1]].
      destruct ov1_src; inv Hov1; clear Hov1.
      - exploit IHe2; eauto. intros [ov2_src [Hov2_src Hov2]].
        destruct ov2_src; inv Hov2; clear Hov2.
        + eexists. split.
          * eapply bindPP_right; eauto. right. split; [eapply sem_exp_type; eauto|].
            eapply bindPP_right; eauto. left. split; eauto.
            intro X. apply H0. inv HP0.
          * constructor.
        + eexists. split; [|constructor].
          eapply bindPP_right; eauto. right. split; [eapply sem_exp_type; eauto|].
          econstructor; eauto.
      - eexists. split; [|constructor].
        econstructor; eauto.
    }
    { exploit IHe1; eauto. intros [ov1_src [Hov1_src Hov1]].
      destruct ov1_src; inv Hov1; clear Hov1.
      - exploit IHe2; eauto. intros [ov2_src [Hov2_src Hov2]].
        destruct ov2_src; inv Hov2; clear Hov2.
        + destruct (sem_bop mem_src bop t t1 t0 t2) eqn:X.
          * destruct (Bop.has_type bop t t0) eqn:Y; [|done].
            exploit mrel_bop; eauto.
            { splits; eauto.
              eapply sem_exp_type; eauto.
              eapply sem_exp_type; eauto.
            }
            intros [v3' [Hv3' Hv3]]. rewrite Hv3'.
            eexists. split; [|apply lift2_Some; eauto].
            eapply bindPP_right; eauto. right. split; [eapply sem_exp_type; eauto|].
            eapply bindPP_right; eauto. right. split; [eapply sem_exp_type; eauto|].
            auto.
          * eexists. split; [|constructor].
            eapply bindPP_right; eauto. right. split; [eapply sem_exp_type; eauto|].
            eapply bindPP_right; eauto. right. split; [eapply sem_exp_type; eauto|].
            auto.
        + eexists. split; [|constructor].
          eapply bindPP_right; eauto. right. split; [eapply sem_exp_type; eauto|].
          econstructor; eauto.
      - eexists. split; [|constructor].
        econstructor; eauto.
    }
Qed.

Lemma mrel_value_addr
      v_src v_tgt mem_src mem_tgt b_src b_tgt blk_src blk_tgt i (mrel:MemRel.t)
      (Hv: MemEquiv.rel_value mrel mem_src mem_tgt v_src v_tgt)
      (Hb: mrel.(MemRel.equiv) b_src b_tgt)
      (Hblk_src: mem_src ! b_src = Some blk_src) (Haddr_src: blk_src.(Mem.Block.address) = None)
      (Hblk_tgt: mem_tgt ! b_tgt = Some blk_tgt) (Haddr_tgt: blk_tgt.(Mem.Block.address) = Some i):
  MemEquiv.rel_value
    mrel
    (PTree.set b_src (Mem.Block.mk
                        blk_src.(Mem.Block.valid)
                        blk_src.(Mem.Block.accessible)
                        blk_src.(Mem.Block.size)
                        (Some i)
                        blk_src.(Mem.Block.content))
               mem_src)
    mem_tgt
    v_src v_tgt.
Proof.
  intros. inv Hv; constructor; auto.
Qed.

Lemma mrel_cast
      v_src v_tgt typ mem_src mem_tgt (mrel:MemRel.t)
      (Hmrel: mrel mem_src mem_tgt)
      (Hv: MemEquiv.rel_value mrel mem_src mem_tgt v_src v_tgt)
      v2_tgt mem2_tgt (Hcast_tgt: sem_cast mem_tgt typ v_tgt (Some (v2_tgt, mem2_tgt))):
  <<Hcast_src: sem_cast mem_src typ v_src None>> \/
  exists v2_src mem2_src (mrel2:MemRel.t),
    <<Hcast_src: sem_cast mem_src typ v_src (Some (v2_src, mem2_src))>> /\
    <<Hmrel2: mrel2 mem2_src mem2_tgt>> /\
    <<Hmrel2_le: MemRel.le_public mrel mrel2>> /\
    <<Hv2: MemEquiv.rel_value mrel2 mem2_src mem2_tgt v2_src v2_tgt>>.
Proof.
  unfold sem_cast.
  intros; destruct typ; tac.
  - inv Hv; clear Hv.
    + right. eexists. eexists. eexists. splits; eauto.
      reflexivity.
      repeat intro. auto.
      repeat intro. auto.
      constructor.
    + exploit MemRel.ptrintcast_public; eauto.
      { inv Hmrel. }
      { inv Hmrel. }
      intros [m2_src [rel2 [[? ?] [Hcast_src [Hrel2 Hrel_le]]]]]. subst.
      right. eexists. eexists. eexists. splits; eauto.
      constructor.
  - inv Hv; clear Hv.
    + destruct (Mem.denormalize i mem_src) as [[? ?]|] eqn:X; [|by left].
      exploit MemRel.denormalize_public; eauto.
      intros [b_tgt [Hb_tgt Hcast]].
      rewrite Hcast in *. inv Hcast_tgt.
      right. eexists. eexists. eexists. splits; eauto.
      reflexivity.
      repeat intro. auto.
      repeat intro. auto.
      constructor. auto.
    + right. eexists. eexists. eexists. splits; eauto.
      reflexivity.
      constructor. auto.
Qed.

Lemma mrel_cast_None
      v_src v_tgt typ mem_src mem_tgt (mrel:MemRel.t)
      (Hmrel: mrel mem_src mem_tgt)
      (Hv: MemEquiv.rel_value mrel mem_src mem_tgt v_src v_tgt)
      (Hcast_tgt: sem_cast mem_tgt typ v_tgt None):
  <<Hcast_src: sem_cast mem_src typ v_src None>>.
Proof.
  unfold sem_cast.
  intros; destruct typ; tac.
  - inv Hv; clear Hv.
    destruct (Mem.normalize ti o mem_tgt) eqn:X; inv Hcast_tgt.
    inv Hr2.
  - inv Hv; clear Hv.
    destruct (Mem.denormalize i mem_src) eqn:Y; [|done].
    destruct p.
    exploit MemRel.denormalize_public; eauto. intros [b_tgt [Hb Hb_tgt]].
    rewrite Hb_tgt in Hcast_tgt. congruence.
Qed.

Ltac rtac :=
    repeat
      (tac;
       try match goal with
         | [H: Vlib.is_true (option_to_bool (match ?cond with Some _ => _ | None => None end)) |- _] =>
           let Hcond := fresh "Hcond" in
           destruct cond eqn:Hcond
         | [H: Vlib.is_true (option_to_bool (match ?cond with Typ.int => None | Typ.ptr => _ end)) |- _] =>
           let Hcond := fresh "Hcond" in
           destruct cond eqn:Hcond
         | [H: Vlib.is_true (option_to_bool (match ?cond with Typ.ptr => None | Typ.int => _ end)) |- _] =>
           let Hcond := fresh "Hcond" in
           destruct cond eqn:Hcond
         | [H: bindPP (sem_exp ?a ?b ?c ?d) _ _ |- _] =>
           inversion H; clear H; tac; exploit mrel_exp; eauto; intros; des
         | [H: ?a = Some _ |- Vlib.is_true (option_to_bool ?a)] =>
           by rewrite H
         | [H: RExp.has_type ?a ?b ?c = Some _ |- Vlib.is_true (option_to_bool (RExp.has_type ?a ?b ?c))] =>
           by rewrite H
       end).

Lemma mrel_rexp
      (rexp:RExp.t) v_tgt evt mem2_tgt
      env_src env_tgt genv_src genv_tgt (mem_src mem_tgt:Mem.t) (mrel:MemRel.t)
      (Hmrel_genv: mrel_genv mrel genv_src genv_tgt)
      (Hmrel_env: mrel_env mrel mem_src mem_tgt env_src env_tgt)
      (Hmrel: mrel mem_src mem_tgt)
      (Henv_src_wf: Env.wf (Env.keys env_src) env_src)
      (Henv_tgt_wf: Env.wf (Env.keys env_tgt) env_tgt)
      (Hrexp_src_wf: RExp.has_type (PTree.keys genv_src) (Env.keys env_src) rexp)
      (Hrexp_tgt_wf: RExp.has_type (PTree.keys genv_tgt) (Env.keys env_tgt) rexp)
      (Hv_tgt: sem_rexp genv_tgt env_tgt mem_tgt rexp (Some (v_tgt, evt, mem2_tgt))):
  <<Herror: sem_rexp genv_src env_src mem_src rexp None>> \/
  exists v_src mem2_src (mrel2:MemRel.t),
    <<Hv_src: sem_rexp genv_src env_src mem_src rexp (Some (v_src, evt, mem2_src))>> /\
    <<Hmrel2: mrel2 mem2_src mem2_tgt>> /\
    <<Hmrel2_le: MemRel.le_public mrel mrel2>> /\
    <<Hv: MemEquiv.rel_value mrel2 mem2_src mem2_tgt v_src v_tgt>>.
Proof.
  destruct rexp; rtac.
  - exploit mrel_exp; eauto; intro; rtac. inv H0.
    { left. econstructor; eauto. }
    right. eexists; eexists; eexists; split.
    eapply bindPP_right; eauto.
    split; eauto; splits; try done; try (unnw; reflexivity).
  - destruct (Exp.has_type (PTree.keys genv_src) (Env.keys env_src) size) eqn: ty; rtac.
    exploit mrel_exp; eauto; rtac.
    intro; rtac. inv H0.
    { left. econstructor; eauto. }
    exploit sem_exp_type; eauto; intro wt.
    destruct a; rtac.
    exploit MemRel.alloc_tgt; eauto; rtac; intro Hmrel'.
    exploit MemRel.alloc_src.
    { apply Hmrel'. }
    { rtac. }
    clear Hmrel'; intro Hmrel'.
    exploit MemRel.public.
    { apply Hmrel'. }
    { rtac; rewrite peq_true; eauto. }
    { rtac; rewrite peq_true; eauto. }
    { rtac.
      constructor; eauto; intros; rtac.
      rewrite ZMap.gi; constructor.
    }
    clear Hmrel'; intro Hmrel'; rtac.
    right. eexists; eexists; eexists; split.
    eapply bindPP_right; eauto; rtac.
    inversion H0; rtac.
    destruct v1; inv Hcond. inv H0. clear H0. inv HP1. clear HP1. rtac.
    split; eauto.
    split.
    { inversion Hmrel; destruct Hpublic; constructor; repeat intro; rtac.
      { constructor; rtac.
        { unfold MemEquiv.add; unfold MemEquiv.to_relation; intros; rtac.
          rewrite peq_false; auto.
          exploit Hpublic; eauto; intro Hblock.
          inversion Hblock; inversion Hsrc_wf.
          exploit _nextb; eauto; apply Plt_ne.
        }
        { apply (Mem.alloc_le i (MemRel.src mrel) (Mem.nextb (MemRel.src mrel)) 
                              {| Mem.nextb := Pos.succ (Mem.nextb (MemRel.src mrel)) |}); auto.
        }
        { apply (Mem.alloc_le i (MemRel.tgt mrel) (Mem.nextb (MemRel.tgt mrel)) 
                              {| Mem.nextb := Pos.succ (Mem.nextb (MemRel.tgt mrel)) |}); auto.
        }
      }
      { assert (i0 <> Mem.nextb (MemRel.src mrel)); rtac.
        inversion Hsrc_wf; exploit _nextb; eauto; apply Plt_ne.
      }
      { assert (i0 <> Mem.nextb (MemRel.tgt mrel)); rtac.
        inversion Htgt_wf; exploit _nextb; eauto; apply Plt_ne.
      }
    }
    rtac.
    exploit (MemEquiv.rel_value_p (MemEquiv.add (Mem.nextb mem_src) (Mem.nextb mem_tgt) (MemRel.equiv mrel))).
    { unfold MemEquiv.add; unfold MemEquiv.to_relation; rtac.
      rewrite peq_true; eauto.
    }
    intro X. splits; try by apply X.
  - destruct (Exp.has_type (PTree.keys genv_src) (Env.keys env_src) exp) eqn: ty; rtac.
    exploit mrel_exp; eauto.
    { by rewrite ty. }
    intro; rtac.
    inv H0; clear H0.
    { left. econstructor; eauto. }
    exploit free_public'; eauto; intro free_cases; destruct free_cases; rtac.
    { left. eapply bindPP_right; eauto. by rewrite H0. }
    right. eexists; eexists; eexists; split.
    eapply bindPP_right; eauto.
    rewrite H0; rtac.
    split; eauto.
    splits; try done; try (constructor; rtac).
  - destruct (Exp.has_type (PTree.keys genv_src) (Env.keys env_src) exp) eqn: ty; rtac.
    exploit mrel_exp; eauto.
    { by rewrite ty. }
    intro; rtac.
    inv H0; clear H0.
    { left. econstructor; eauto. }
    exploit mrel_cast; eauto; intro; rtac. destruct x0.
    { left. eapply bindPP_right; eauto. constructor; auto. }
    des. right. eexists. eexists. eexists. splits; eauto.
    eapply bindPP_right; eauto.
    eapply bindPP_right; eauto.
  - setoid_rewrite bindPP_True.
    right. eexists; eexists; eexists; split; eauto; split; eauto; splits; try done; [unnw; reflexivity | constructor].
  - destruct (Exp.has_type (PTree.keys genv_src) (Env.keys env_src) exp) eqn: ty; rtac.
    exploit mrel_exp; eauto.
    { by rewrite ty. }
    intro; rtac.
    inv H0; clear H0.
    { left. econstructor; eauto. }
    exploit sem_exp_type; eauto; intro wt.
    destruct a; rtac.
    inversion HP; rtac.
    (* unfold Mem.val2int in Hcond; rewrite <- H3 in Hcond; rtac. *)
    right. eexists; eexists; eexists; split.
    eapply bindPP_right; eauto; rtac.
    split; eauto.
    splits; try done. reflexivity.
Qed.

Lemma mrel_rexp_None
      (rexp:RExp.t)
      env_src env_tgt genv_src genv_tgt (mem_src mem_tgt:Mem.t) (mrel:MemRel.t)
      (Hmrel_genv: mrel_genv mrel genv_src genv_tgt)
      (Hmrel_env: mrel_env mrel mem_src mem_tgt env_src env_tgt)
      (Hmrel: mrel mem_src mem_tgt)
      (Henv_src_wf: Env.wf (Env.keys env_src) env_src)
      (Henv_tgt_wf: Env.wf (Env.keys env_tgt) env_tgt)
      (Hrexp_src_wf: RExp.has_type (PTree.keys genv_src) (Env.keys env_src) rexp)
      (Hrexp_tgt_wf: RExp.has_type (PTree.keys genv_tgt) (Env.keys env_tgt) rexp)
      (Hv_tgt: sem_rexp genv_tgt env_tgt mem_tgt rexp None):
  sem_rexp genv_src env_src mem_src rexp None.
Proof.
  destruct rexp; rtac.
  - inv Hov. econstructor; eauto.
  - inv Hov. econstructor; eauto.
  - inv Hov; clear Hov; [econstructor; eauto|inv HP].
    eapply bindPP_right; eauto.
  - inv Hov. econstructor; eauto.
  - inv Hov; clear Hov; [econstructor; eauto|inv HP].
    eapply bindPP_right; eauto.
    eapply bindPP_right; eauto.
    destruct (Mem.free (Value.ptr si o) mem_src) eqn:H1_src; rtac.
    exploit MemRel.free_public; eauto. intro. des.
    rewrite x0 in Hr2. inv Hr2.
  - inv Hov. econstructor; eauto.
  - inv Hov; clear Hov; [econstructor; eauto|inv HP].
    eapply bindPP_right; eauto.
    inv Hr2. exploit mrel_cast_None; eauto. intro X.
    econstructor; eauto.
    eapply bindPP_right; eauto.
    inv Hr2. exploit mrel_cast_None; eauto. intro X.
    econstructor; eauto.
  - inv Hv_tgt.
  - inv Hov. econstructor; eauto.
  - inv Hov; clear Hov; [econstructor; eauto|inv HP].
    eapply bindPP_right; eauto.
Qed.

Lemma mrel_instr
      (istr:Instr.t) istrs
      (st_src st_tgt:State)
      (mrel:MemRel.t)
      (sst_src sst_tgt:StaticState)
      event st2_tgt
      (Hncall: Instr.is_call istr = false)
      (Hinstrs_src: st_src.(context).(instrs) = istr::istrs)
      (Hinstrs_tgt: st_tgt.(context).(instrs) = istr::istrs)
      (Hmrel_genv: mrel_genv mrel sst_src.(genv) sst_tgt.(genv))
      (Hmrel_env: mrel_env mrel st_src.(mem) st_tgt.(mem) st_src.(context).(lenv) st_tgt.(context).(lenv))
      (Hmrel: mrel st_src.(mem) st_tgt.(mem))
      (Hwf_src: Context_wf sst_src st_src.(context))
      (Hwf_tgt: Context_wf sst_tgt st_tgt.(context))
      (Htgt: sem_instr sst_tgt st_tgt (Some (event, st2_tgt))):
  <<Herror: sem_instr sst_src st_src None>> \/
  exists st2_src (mrel2:MemRel.t),
    <<Hst2_src: sem_instr sst_src st_src (Some (event, st2_src))>> /\
    <<Histrs: st2_src.(context).(instrs) = st2_tgt.(context).(instrs)>> /\
    <<Hmrel2: mrel2 st2_src.(mem) st2_tgt.(mem)>> /\
    <<Hmrel2_le: MemRel.le_public mrel mrel2>> /\
    <<Hmrel_env: mrel_env mrel2 st2_src.(mem) st2_tgt.(mem) st2_src.(context).(lenv) st2_tgt.(context).(lenv)>>.
Proof.
  unfold sem_instr in Htgt. destruct istr; tac.
  { inv Hwf_src; inv Hinstrs. rtac.
    inv Hwf_tgt; inv Hinstrs. rtac.
    unfold sem_instr in *. rtac.
    exploit mrel_Env_init; eauto. intro. rtac.
    right. eexists. exists mrel. splits; rtac.
    - eapply bindPP_right; eauto.
    - rtac.
    - rtac.
    - reflexivity.
    - rtac.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    unfold sem_instr. rtac.
    destruct var; rtac.
    - destruct Hr2; rtac.
      exploit mrel_rexp; eauto; rtac.
      intros [X|X]; [left|right]; des.
      { by apply bindPP_left_None. }
      eexists. exists mrel2. splits; eauto.
      + eapply bindPP_right; eauto.
        right. split; eauto. rtac.
        exploit (sem_rexp_type (genv sst_src)); eauto.
      + rtac.
      + rtac.
      + rtac. apply mrel_Env_set; rtac.
        eapply (mrel_env_le lenv1 lenv0 mem1); eauto.
        apply MemRel.le_public_le. eauto.
    - exploit mrel_rexp; eauto; rtac.
      intros [X|X]; [left|right]; des.
      { by apply bindPP_left_None. }
      eexists. exists mrel2. splits; eauto.
      + eapply bindPP_right; eauto.
      + rtac.
      + rtac.
      + rtac. eapply (mrel_env_le lenv1 lenv mem1); eauto.
        apply MemRel.le_public_le. eauto.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    unfold sem_instr. rtac.
    destruct var; rtac.
    - destruct Hr2; rtac.
      exploit (mrel_exp rhs); eauto; rtac.
      intro. des. inv Hov; clear Hov.
      { left. econstructor; eauto. }
      destruct (Mem.load a mem1) eqn:Hload_src; rtac.
      { exploit MemRel.load_public; eauto. intro. rtac.
        rewrite Hcond in H0. inv H0. clear H0.
        destruct (Value.has_type t1 (fst t)) eqn:H1_src.
        Focus 2.
        { left. eapply bindPP_right; eauto.
          rewrite Hload_src. rtac.
          rewrite H1_src. by left.
        }
        Unfocus.
        right. eexists. exists mrel. splits; eauto.
        + eapply bindPP_right; eauto. rewrite Hload_src. rtac.
          right. split; eauto.
        + rtac.
        + rtac.
        + reflexivity.
        + rtac. apply mrel_Env_set; rtac.
      }
      { left. unfold sem_instr. rtac.
        eapply bindPP_right; eauto. by rewrite Hload_src.
      }
    - exploit (mrel_exp rhs); eauto; rtac.
      intro. des. inv Hov; clear Hov.
      { left. econstructor; eauto. }
      destruct (Mem.load a mem1) eqn:Hload_src; rtac.
      { exploit MemRel.load_public; eauto. intro. rtac.
        right. eexists. exists mrel. splits; eauto.
        + eapply bindPP_right; eauto. rewrite Hload_src. rtac.
        + rtac.
        + rtac.
        + reflexivity.
        + rtac.
      }
      { left. unfold sem_instr. rtac.
        eapply bindPP_right; eauto. by rewrite Hload_src.
      }
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    unfold sem_instr. rtac.
    exploit (mrel_exp lhs); eauto; rtac.
    intro. des. inv Hov.
    { left. econstructor; eauto. }
    exploit (mrel_exp rhs); eauto; rtac.
    intro. des. inv Hov0.
    { left. eapply bindPP_right; eauto. econstructor; eauto. }
    exploit (store_public' mrel mem1 mem0 Hmrel a v1); eauto. intro. des.
    { left. unfold sem_instr. rtac.
      repeat (eapply bindPP_right; eauto). by rewrite x0.
    }
    right. eexists. eexists. splits.
    - repeat (eapply bindPP_right; eauto).
      rewrite x0. rtac.
    - rtac.
    - eauto.
    - constructor; eauto; rtac.
    - rtac.
      repeat intro. exploit mrel_env_le; eauto.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    unfold sem_instr. rtac.
    exploit (mrel_exp cond); eauto; rtac. intro. des. inv Hov. 
    { left. econstructor; eauto. }
    unfold Mem.val2bool in *. rtac.
    destruct (Mem.val2int a mem1) eqn: Hint_tgt; rtac.
    { exploit MemRel.val2int_public; eauto. intro X. rewrite X in Hcond0. inv Hcond0.
      right. eexists. exists mrel. splits.
      - repeat (eapply bindPP_right; eauto).
        rewrite Hint_tgt. rtac.
      - rtac.
      - rtac.
      - reflexivity.
      - rtac.
    }
    { left. unfold sem_instr. rtac.
      repeat (eapply bindPP_right; eauto).
      unfold Mem.val2bool.
      by rewrite Hint_tgt.
    }
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    unfold sem_instr. rtac.
    exploit (mrel_exp cond); eauto; rtac. intro. des. inv Hov.
    { left. econstructor; eauto. }
    unfold Mem.val2bool in *. rtac.
    destruct (Mem.val2int a mem1) eqn: Hint_tgt; rtac.
    { exploit MemRel.val2int_public; eauto. intro X. rewrite X in Hcond0. inv Hcond0.
      right. eexists. exists mrel. splits.
      - repeat (eapply bindPP_right; eauto).
        rewrite Hint_tgt. rtac.
      - rtac.
      - rtac.
      - reflexivity.
      - rtac.
    }
    { left. unfold sem_instr. rtac.
      repeat (eapply bindPP_right; eauto).
      unfold Mem.val2bool.
      by rewrite Hint_tgt.
    }
  }
Qed.

Lemma mrel_instr_None
      (istr:Instr.t) istrs_src istrs_tgt
      (st_src st_tgt:State)
      (mrel:MemRel.t)
      (sst_src sst_tgt:StaticState)
      (Hncall: Instr.is_call istr = false)
      (Hinstrs_src: st_src.(context).(instrs) = istr::istrs_src)
      (Hinstrs_tgt: st_tgt.(context).(instrs) = istr::istrs_tgt)
      (Hmrel_genv: mrel_genv mrel sst_src.(genv) sst_tgt.(genv))
      (Hmrel_env: mrel_env mrel st_src.(mem) st_tgt.(mem) st_src.(context).(lenv) st_tgt.(context).(lenv))
      (Hmrel: mrel st_src.(mem) st_tgt.(mem))
      (Hwf_src: Context_wf sst_src st_src.(context))
      (Hwf_tgt: Context_wf sst_tgt st_tgt.(context))
      (Htgt: sem_instr sst_tgt st_tgt None):
  sem_instr sst_src st_src None.
Proof.
  unfold sem_instr in Htgt. destruct istr; tac.
  { inv Hwf_src; inv Hinstrs. rtac.
    inv Hwf_tgt; inv Hinstrs. rtac.
    inv Htgt.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    destruct var; rtac.
    - inv Htgt; rtac.
      + exploit mrel_rexp_None; eauto; rtac. intro X.
        unfold sem_instr. rtac. by apply bindPP_left_None.
      + destruct Hr2; rtac.
        unfold sem_instr. rtac.
        exploit mrel_rexp; eauto; rtac. intros [X|X]; des.
        { by apply bindPP_left_None. }
        eapply bindPP_right; eauto.
        left. split; rtac. intro X. apply H.
        exploit (sem_rexp_type (genv sst_tgt)); eauto.
    - inv Htgt; rtac.
      exploit mrel_rexp_None; eauto; rtac. intro X.
      unfold sem_instr. rtac. by apply bindPP_left_None.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    unfold sem_instr. rtac.
    destruct var; rtac.
    - inv Hov. econstructor; eauto.
    - inv Hov.
      { econstructor; eauto. }
      eapply bindPP_right; eauto.
      destruct (Mem.load a mem0) eqn:H1_src; rtac.
      exploit MemRel.load_public; eauto. intro. rtac.
      rewrite H in Hr2. rtac. destruct Hr2; rtac.
      left. split; rtac. intro X. apply H1.
      exploit mrel_has_type; eauto.
    - inv Hov. econstructor; eauto.
    - inv Hov.
      { econstructor; eauto. }
      eapply bindPP_right; eauto.
      destruct (Mem.load a mem0) eqn:H1_src; rtac.
      exploit MemRel.load_public; eauto. intro. rtac.
      rewrite H in Hr2. rtac.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. tac.
    inv Htgt; clear Htgt.
    { exploit mrel_exp; try apply Hr1; eauto. by rewrite e0.
      intros [ov_src [Hov_src Hov]]. inv Hov.
      econstructor; eauto.
    }
    exploit (mrel_exp lhs); eauto.
    { by rewrite e0. }
    intro. tac. inv H3.
    { econstructor; eauto. }
    inv Hr2; clear Hr2; tac.
    { exploit mrel_exp; try apply Hr1; eauto. by rewrite e0.
      intros [ov_src [Hov_src Hov]]. inv Hov.
      { econstructor; eauto. }
      exploit mrel_exp; try apply Hr0; eauto.
      intros [ov_src' [Hov_src' Hov']]. inv Hov'.
      eapply bindPP_right; simpl; eauto.
      econstructor; eauto.
    }
    exploit (mrel_exp rhs); eauto.
    intro. tac. inv H6.
    { eapply bindPP_right; eauto. simpl.
      econstructor; eauto.
    }
    unfold sem_instr. tac.
    repeat (eapply bindPP_right; eauto).
    destruct (Mem.store a a0 mem0) eqn:Hstore_src; tac.
    exploit (MemRel.store_public mrel mem0 mem Hmrel a v1); eauto.
    intro. tac.
    rewrite H7 in Hr3. tac.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. tac.
    inv Htgt; clear Htgt.
    { exploit mrel_exp; try apply Hr1; eauto. by rewrite e0.
      intros [ov_src [Hov_src Hov]]. inv Hov.
      econstructor; eauto.
    }
    exploit mrel_exp; eauto.
    { by rewrite e0. }
    intro. tac. inv H3.
    { econstructor; eauto. }
    unfold sem_instr. tac.
    eapply bindPP_right; eauto.
    unfold Mem.val2bool in *.
    destruct (Mem.val2int a mem0) eqn:Hint_src; tac.
    exploit MemRel.val2int_public; eauto. intro X. rewrite X in Hr2. tac.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. tac.
    inv Htgt; clear Htgt.
    { exploit mrel_exp; try apply Hr1; eauto. by rewrite e0.
      intros [ov_src [Hov_src Hov]]. inv Hov.
      econstructor; eauto.
    }
    exploit mrel_exp; eauto.
    { by rewrite e0. }
    intro. tac. inv H6.
    { econstructor; eauto. }
    unfold sem_instr. tac.
    eapply bindPP_right; eauto.
    unfold Mem.val2bool in *.
    destruct (Mem.val2int a mem0) eqn:Hint_src; tac.
    exploit MemRel.val2int_public; eauto. intro X. rewrite X in Hr2. tac.
  }
Qed.
